# -*- coding: utf-8 -*-
"""
This module contains methods related to reading input files, such as targets.csv, probes.csv etc.
"""

import re
import sys
import os, hashlib
import pandas as pd
import Bio.Seq

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
sh = logging.StreamHandler()
sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
log.addHandler(sh)

mipgen_columns = ['chr','ext_probe_sequence','lig_probe_sequence','mip_scan_start_position','mip_scan_stop_position','probe_strand','mip_name']
heatseq_columns = ['chromosome','ext_sequence','lig_sequence','target_start','target_stop','probe_strand','probe_id']
probe_columns = ['id', 'first_primer_5to3', 'second_primer_5to3', 'chr', 'target_start', 'target_end', 'strand']

class AmplimapReaderException(Exception):
    """
    Will be raised if reading one of the standard input files has failed, with the aim of providing a more useful error message in the user-visible output.
    """
    def __init__(self, e: Exception, filename: str, should_have_header: bool = None):
        if isinstance(e, ValueError):
            err = 'Invalid value. This usually indicates that there were non-numeric characters (eg. letters, spaces, ...) in a column that should only contain numbers.\n\nMessage: {}'.format(
                str(e)
            )
        elif isinstance(e, AssertionError):
            err = 'Assertion error. This usually indicates that one of the values is invalid or in an unexpected range.\n\nMessage: {}'.format(
                str(e)
            )
        else:
            err = str(e)

        header_note = ''
        if should_have_header is not None:
            if should_have_header:
                header_note = 'Please note that this file SHOULD start with a header line specifying the column names.'
            else:
                header_note = 'Please note that this file SHOULD NOT contain any header line.'

        #add better message to exception
        message = '\n\nError while reading file {}:\n{}\n\n{}\n\n'.format(
            filename,
            err,
            header_note
        )

        #init base exception
        super().__init__(message)

def get_code_versions(path: str = '.') -> dict:
    """
    Get the file modification times of common code files for versioning.
    """
    VERSIONS = {}
    VERSIONS['_amplimap'] = str(__version__)
    VERSIONS['Snakefile'] = str(os.path.getmtime(os.path.join(path, 'Snakefile')))
    VERSIONS['parse_reads'] = str(os.path.getmtime(os.path.join(path, 'parse_reads.py')))
    VERSIONS['pileup'] = str(os.path.getmtime(os.path.join(path, 'pileup.py')))
    VERSIONS['stats_alignment'] = str(os.path.getmtime(os.path.join(path, 'stats_alignment.py')))
    VERSIONS['coverage'] = str(os.path.getmtime(os.path.join(path, 'coverage.py')))
    VERSIONS['simulate'] = str(os.path.getmtime(os.path.join(path, 'simulate.py')))
    VERSIONS['variants'] = str(os.path.getmtime(os.path.join(path, 'variants.py')))
    return VERSIONS

def get_file_hashes(path: str = '.') -> dict:
    """
    Get SHA256 hashes for common input files, so we can make sure these didn't change between runs.
    """
    hashes = {}
    for fn in ['probes.csv', 'targets.bed', 'targets.csv', 'snps.txt']:
        fpath = os.path.join(path, fn)
        if os.path.isfile(fpath):
            file_hash = hashlib.sha256()
            with open(fpath, 'rb', buffering = 0) as file:
                for chunk in iter(lambda : file.read(128*1024), b''):
                    file_hash.update(chunk)
            hashes[fn] = file_hash.hexdigest()
        else:
            hashes[fn] = ''

    return hashes

def merge_probes_by_id(rows: pd.DataFrame) -> pd.Series:
    """
    Merge multiple probes.csv rows with the same probe ID together, to handle cases where MIPGEN generated multiple version because of a SNP.
    """
    #grab first row as template (use .loc instead of .iloc?)
    first_row = rows.iloc[0].copy()

    first_row['n_merged'] = len(rows)
    first_row['n_wildcards'] = None

    if len(rows) > 1:
        first_row['n_wildcards'] = 0

        #make sure these are all the same
        assert (rows['chr'] == first_row['chr']).all(), 'found multiple rows with same probe name but different chr: %s' % first_row['id']
        assert (rows['target_start'] == first_row['target_start']).all(), 'found multiple rows with same probe name but different target_start: %s' % first_row['id']
        assert (rows['target_end'] == first_row['target_end']).all(), 'found multiple rows with same probe name but different target_end: %s' % first_row['id']
        assert (rows['strand'] == first_row['strand']).all(), 'found multiple rows with same probe name but different strand: %s' % first_row['id']

        #merge primers together
        for primer_col in ['first_primer_5to3', 'second_primer_5to3']:
            assert (rows[primer_col].str.len() == len(first_row[primer_col])).all(), 'found multiple rows with same probe id but primer sequences were different lengths'
            for ixrow in range(1, len(rows)):
                for ixchar in range(len(first_row[primer_col])):
                    #if this row doesn't match the first row's sequence
                    if first_row[primer_col][ixchar] != rows.iloc[ixrow][primer_col][ixchar]:
                        #replace character by dot, but only if it hasn't been replaced yet
                        if first_row[primer_col][ixchar] != '.':
                            first_row['n_wildcards'] += 1
                            first_row[primer_col] = first_row[primer_col][:ixchar] + '.' + first_row[primer_col][(ixchar+1):]

        #make sure we actually replaced something with a wildcard
        assert first_row['n_wildcards'] > 0, 'found multiple rows with same probe name but primer sequences were identical: %s' % first_row['id']
        assert first_row['n_wildcards'] < 10, 'found too many differences between primer sequences: %s' % first_row['id']

    return first_row

def read_and_convert_mipgen_probes(path: str, sep=',') -> pd.DataFrame:
    """
    Read probes file from MIPGEN in CSV or TSV format and generate an amplimap probes.csv from it.
    """
    try:
        design = pd.read_csv(path, sep=sep)

        log.info('Read mipgen table from %s -- found %d probes', path, len(design))
        original_len = len(design)
        design.dropna(axis=0, how='all', inplace=True)
        if original_len != len(design):
            log.info('%d probes after dropping empty rows', len(design))
        assert len(design) > 0, 'mipgen table is empty!'

        # make sure we have all the columns we expect
        for column in mipgen_columns:
            if not column in design.columns:
                raise Exception('invalid mipgen table: column %s is missing!' % column)

        # rename columns
        design.rename(columns = {'mip_scan_start_position': 'target_start', 'mip_scan_stop_position': 'target_end', 'probe_strand': 'strand', 'mip_name': 'id'}, inplace=True)

        # in smMIPs, the first arm (read1) seems to be the reverse primer, which would normally be the second arm
        # its sequence in the table seems to be on the forward strand, which needs to be RC'd to get the actual read sequence (5p->3p)
        design['first_primer_5to3'] = [str(Bio.Seq.Seq(s).reverse_complement()) for s in design['lig_probe_sequence'].str.upper()]
        design['second_primer_5to3'] = design['ext_probe_sequence'].str.upper()

        # if we have a probe with a SNP in it we'll get multiple probes with NAME_SNP_a, NAME_SNP_b, etc
        # here we will figure out the normal id, which we then use for grouping
        design['id_parts'] = design['id'].str.split('_SNP_')
        design['id'] = [row.id_parts[0] if len(row.id_parts) == 2 else row.id for row in design.itertuples()]
        # design['snp_index'] = [row.id_parts[1] if len(row.id_parts) == 2 else None for row in design.itertuples()]

        # find probes with duplicate ID and replace their different bases by dots
        original_len = len(design)
        design = design.groupby('id', as_index=False).apply(merge_probes_by_id)
        if original_len != len(design):
            log.info('%d/%d probes left after merging probes by ID', len(design), original_len)

        # reorder and extract columns
        design = design[probe_columns]
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = True).with_traceback(sys.exc_info()[2])

    return design

def read_and_convert_heatseq_probes(path: str) -> pd.DataFrame:
    """
    UNTESTED: Read probes file from Roche heatseq in CSV format and generate an amplimap probes.csv from it.
    """
    try:
        design = pd.read_csv(path, sep='\t', comment='#')

        log.info('Read heatseq table from %s -- found %d probes', path, len(design))
        original_len = len(design)
        design.dropna(axis=0, how='all', inplace=True)
        if original_len != len(design):
            log.info('%d probes after dropping empty rows', len(design))
        assert len(design) > 0, 'heatseq table is empty!'

        #make sure we have all the columns we expect
        for column in heatseq_columns:
            if not column in design.columns:
                raise Exception('invalid heatseq table: column %s is missing!' % column)

        #rename columns
        design.rename(columns = {'chromosome': 'chr', 'target_stop': 'target_end', 'probe_strand': 'strand', 'probe_id': 'id'}, inplace=True)

        #In this protocol, read1 should contain extension primer and read2 the ligation primer
        #Note that this is the opposite of smMIPs!
        design['second_primer_5to3'] = [str(Bio.Seq.Seq(s).reverse_complement()) for s in design['lig_sequence'].str.upper()]
        design['first_primer_5to3'] = design['ext_sequence'].str.upper()

        #detect SNPs and replace them by dot
        original_len = len(design)
        design = design.groupby('id').apply(merge_probes_by_id)
        if original_len != len(design):
            log.info('%d/%d probes left after merging probes by ID', len(design), original_len)
        #reset index to turn id into column again
        design.reset_index(inplace=True)

        #reorder and extract columns
        design = design[probe_columns]
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = True).with_traceback(sys.exc_info()[2])

    return design

def read_new_probe_design(path: str, reference_type: str = 'genome') -> pd.DataFrame:
    """
    Read amplimap probes.csv file and return pandas dataframe.
    """
    try:
        design = pd.read_csv(path)

        log.info('Read probe design table from %s -- found %d probes', path, len(design))
        if list(design.columns) == mipgen_columns:
            #NB: smmip data seems to be in F2R1 orientation (second read = fwd in genomic coordinates) for fwd probes
            #but F1R2 orientation (second read = rev) for rev probes.
            #cs-tag data seems to be in F1R2 orientation for fwd targets. unclear for rev targets, but presumably F2R1?
            #in other words, CS-tag is in gene orientation, while smMIP is in opposite
            #so both are swapped for MIPs.
            #is this why sequences in probes.csv are currently so confusing?
            log.info('Detected old MIPGEN format, converting...')

            #read the probes file again in old mipgen format and convert
            design = read_and_convert_mipgen_probes(path)

        design = process_probe_design(design, reference_type)
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = True).with_traceback(sys.exc_info()[2])

    return design

def process_probe_design(design: pd.DataFrame, reference_type: str = 'genome') -> pd.DataFrame:
    """
    Read amplimap probes.csv file and return pandas dataframe.
    """
    #now drop empty rows
    original_len = len(design)
    design.dropna(axis=0, how='all', inplace=True)
    if original_len != len(design):
        log.info('%d probes after dropping empty rows', len(design))
    assert len(design) > 0, 'Probe design table is empty!'

    #make sure we have all the columns we expect
    for column in probe_columns:
        if not column in design.columns:
            raise Exception('invalid probe design table: column %s is missing!' % column)

    #fix data types
    design = design.astype(
        {
            'id': str,
            'chr': str,
            'strand': str,
            'first_primer_5to3': str,
            'second_primer_5to3': str,
            #these need to be int instead of uint, otherwise they'll get turned into floats when subtracting an int (as we do below)
            'target_start': int,
            'target_end': int,
        },
        errors = 'raise'
    )

    #calculate some additional columns
    design.loc[design['strand'] == '+', 'probe_start'] = design.loc[design['strand'] == '+', 'target_start'] - design.loc[design['strand'] == '+', 'first_primer_5to3'].str.len()
    design.loc[design['strand'] == '-', 'probe_start'] = design.loc[design['strand'] == '-', 'target_start'] - design.loc[design['strand'] == '-', 'second_primer_5to3'].str.len()
    design.loc[design['strand'] == '+', 'probe_end'] = design.loc[design['strand'] == '+', 'target_end'] + design.loc[design['strand'] == '+', 'second_primer_5to3'].str.len()
    design.loc[design['strand'] == '-', 'probe_end'] = design.loc[design['strand'] == '-', 'target_end'] + design.loc[design['strand'] == '-', 'first_primer_5to3'].str.len()

    design = design.astype(
        {
            'probe_start': int,
            'probe_end': int,
        },
        errors = 'raise'
    )

    design['target_start_0'] = design['target_start'] - 1   # we have 1-based coords now
    design['probe_start_0'] = design['probe_start'] - 1   # we have 1-based coords now
    design['capture_size'] = design['probe_end'] - design['probe_start_0']
    design['target_length'] = design['target_end'] - design['target_start_0']

    invalid_coords = ~(design['target_length'] > 0)
    if invalid_coords.any():
        print(design.loc[invalid_coords, ['id', 'chr', 'probe_start', 'probe_end', 'strand']])
        raise Exception('Encountered probe regions with invalid target coordinates. Please check the input file and the output above. Note that the probe_end position must always be greater than the probe_start.')

    # #chr should start with chr
    # if reference_type == 'genome':
    #     design.loc[~design['chr'].str.startswith('chr'), 'chr'] = ['chr' + c for c in design.loc[~design['chr'].str.startswith('chr'), 'chr']]

    #check that probe ids don't contain unexpected characters (space breaks bam files for example, since umi will be added as tag)
    assert design['id'].str.contains(r'^[a-zA-Z0-9/._:+-]*$').all(), 'probe names may only contain alphanumeric characters and /._:+-'

    #check for dupes
    any_dup = design.duplicated(subset = ['chr', 'probe_start', 'probe_end', 'strand'], keep = False)
    if any_dup.any():
        print('These lines in the design are duplicated:')
        print(design.loc[any_dup, ['chr', 'probe_start', 'probe_end', 'id', 'strand']])

        dup = design.duplicated(subset = ['chr', 'probe_start', 'probe_end', 'strand'], keep = 'first')
        log.warn('Dropping %d/%d duplicated rows', dup.sum(), len(design))
        design = design[~dup]
        log.warn('Now have %d rows after dropping duplicates', len(design))

        raise Exception('Stopping because of duplicate rows in probes.csv!')

    #make sure probes are uppercase
    design['first_primer_5to3'] = design['first_primer_5to3'].str.upper()
    design['second_primer_5to3'] = design['second_primer_5to3'].str.upper()

    #check strand
    if 'strand' in design.columns:
        assert (design['strand'].isin(['+', '-'])).all(), 'strand must be + or -'

    #get reverse complement of sequence for smart trimming and matching to arms reverse read
    for c in ['first_primer_5to3', 'second_primer_5to3']:
        design['%s__rc' % c] = [str(Bio.Seq.Seq(s).reverse_complement()) for s in design[c]]

    #set the index to the probe id (useful for probe stats)
    design.set_index('id', drop = False, inplace = True, verify_integrity = True)

    return design

def read_sample_info(path: str) -> pd.DataFrame:
    """
    Read amplimap sample_info.csv file and return pandas dataframe.
    """
    try:
        sample_infos = pd.read_csv(path, dtype = 'object')

        assert not sample_infos.columns.str.contains('^Unnamed').any(), \
            'Missing column names in sample info table! Please provide names for all columns in the first row.'

        log.info('Read sample info table from %s -- found %d rows', path, len(sample_infos))
        sample_infos.dropna(axis=0, how='all', inplace=True)
        log.info('%d after dropping empty rows', len(sample_infos))

        assert sample_infos.columns[0] == 'Sample' and sample_infos.columns[1] == 'Targets', \
            'Missing Sample/Targets columns! Got %d columns: %s' % (len(sample_infos.columns), '/'.join(sample_infos.columns))

        #split up Targets column into list
        sample_infos['Target'] = sample_infos['Targets'].str.split(';')

        #make new data frame by stacking the Target lists
        #inspired by: https://stackoverflow.com/questions/38372016/split-nested-array-values-from-pandas-dataframe-cell-over-multiple-rows
        non_target_columns = [c for c in sample_infos.columns if c != 'Target']
        sample_infos.set_index(non_target_columns, inplace = True) #this needs need to be all unstacked columns
        #stack each column (only Target actually)
        sample_infos = sample_infos.apply(lambda col: col.apply(pd.Series).stack(), axis=0)
        #turn multilevel index back into columns
        sample_infos.reset_index(inplace = True)
        #drop Targets column and target index column generated by reset_index
        sample_infos.drop(['Targets', 'level_%d' % len(non_target_columns)], axis=1, inplace = True)

        #check for dupes
        if sample_infos.duplicated(['Sample', 'Target']).any():
            print(targets[sample_infos.duplicated(['Sample', 'Target'], keep=False)])
            raise Exception('Invalid sample_infos table -- found duplicated Sample/Target IDs!')

        #set proper index
        sample_infos.set_index(['Sample', 'Target'], drop = True, inplace = True, verify_integrity = True)

        print(sample_infos.head(3))
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = True).with_traceback(sys.exc_info()[2])

    return sample_infos

def read_targets(path: str, check_overlaps: bool = False, reference_type: str = 'genome', file_type: str = 'bed') -> pd.DataFrame:
    """
    Read amplimap targets.csv or targets.bed file and return pandas dataframe.
    """
    try:
        if file_type == 'bed':
            skiprows = 0
            with open(path, 'r') as f:
                for line in f:
                    if line.startswith('track') or line.startswith('browser'):
                        skiprows += 1
                    else:
                        break

            if skiprows > 0:
                log.info('Skipping %d browser/track rows from target BED files', skiprows)

            targets = pd.read_csv(path, header=None, skiprows=skiprows, sep='\t')
            log.info('Read targets from %s: %d rows', path, len(targets))

            if targets.shape[1] == 6:
                targets.columns = ['chr', 'start_0', 'end', 'id', 'score', 'strand']
                #targets = targets[['chr', 'start_0', 'end', 'id']] #remove score/strand
            elif targets.shape[1] == 4:
                targets.columns = ['chr', 'start_0', 'end', 'id']
            elif targets.shape[1] == 3:
                targets.columns = ['chr', 'start_0', 'end']
                targets['id'] = ['target_%d' % i for i in range(len(targets))]
            else:
                print(targets)
                raise Exception('Invalid target BED file: expected to find either 6, 4 or 3 columns separated by tabs.')
        elif file_type == 'csv':
            targets = pd.read_csv(path)

            #check columns
            if not ('chr' in targets.columns and 'start' in targets.columns and 'end' in targets.columns):
                print(targets)
                raise Exception('Invalid target CSV: first line should contain column names, which must include "chr", "start" and "end".')

            #add id if need be
            if not 'id' in targets.columns:
                targets['id'] = ['target_%d' % i for i in range(len(targets))]

            #fix start and calculate start_0
            targets['start'] = targets['start'].astype(int)
            targets['start_0'] = targets['start'] - 1
        else:
            raise Exception('invalid target file type')

        #fix data types
        targets = targets.astype(
            {
                'id': str,
                'chr': str,
                'start_0': int,
                'end': int,
            },
            errors = 'raise'
        )

        #make sure we have no duplicated ids
        if targets.duplicated('id').any():
            print(targets[targets.duplicated('id', keep=False)])
            raise Exception('Invalid target table: found duplicated target ID(s)!')

        #check for overlaps
        if check_overlaps:
            targets_dict = targets.to_dict()
            for ixrow1 in range(len(targets_dict['chr'])-1):
                for ixrow2 in range(ixrow1+1, len(targets_dict['chr'])):
                    if targets_dict['chr'][ixrow1] == targets_dict['chr'][ixrow2] \
                    and targets_dict['start_0'][ixrow1] <= targets_dict['end'][ixrow2] \
                    and targets_dict['end'][ixrow1] >= targets_dict['start_0'][ixrow2]:
                        print(targets.iloc[ixrow1])
                        print(targets.iloc[ixrow2])
                        raise Exception(
                            'Invalid target table: target regions %s and %s overlap! Please merge them or shorten one of the regions' % (
                                targets_dict['id'][ixrow1],
                                targets_dict['id'][ixrow2]
                            )
                        )

        #chr should always start with chr
        targets['chr_original'] = targets['chr']
        # if reference_type == 'genome':
        #     targets.loc[~targets['chr'].str.startswith('chr'), 'chr'] = ['chr' + c for c in targets.loc[~targets['chr'].str.startswith('chr'), 'chr']]

        targets['length'] = targets['end'] - targets['start_0']

        targets_wrong_length = targets['length'] <= 0
        if targets_wrong_length.any():
            print(targets[targets_wrong_length])
            raise Exception('Invalid target table: found targets with length <= 0! Please check start and end coordinates.')

        targets['type'] = 'target'
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = file_type == 'csv').with_traceback(sys.exc_info()[2])

    return targets

def write_targets_bed(path: str, targets: pd.DataFrame):
    """
    Write targets dataframe to bed file.
    """
    if not 'score' in targets.columns:
        targets['score'] = '.'
    if not 'strand' in targets.columns:
        targets['strand'] = '.'

    targets[['chr_original', 'start_0', 'end', 'id', 'score', 'strand']].to_csv(path, header = False, index = False, sep = '\t')

def read_snps_txt(path: str, reference_type: str = 'genome') -> pd.DataFrame:
    """
    Read amplimap snps.txt file and return pandas dataframe.
    """
    try:
        log.info('Loading SNPs from %s', path)
        snps = pd.read_csv(path, header=None, sep='\t')  # pos is 1-based!

        #detect potential issues with spaces instead of tabs
        if snps.shape[1] == 1:
            if ' ' in snps.iloc[0, 0]:
                log.info('SNP table does not seem to be tab-separated, trying again with spaces')
                snps = pd.read_csv(path, header=None, sep=r'\s+')

        if snps.shape[1] == 5:
            snps.columns = ['chr', 'pos', 'id', 'snp_ref', 'snp_alt']
        elif snps.shape[1] == 4:
            snps.columns = ['chr', 'pos', 'id', 'snp_ref']
        elif snps.shape[1] == 3:
            snps.columns = ['chr', 'pos', 'id']
        elif snps.shape[1] == 2:
            snps.columns = ['chr', 'pos']
            snps['id'] = ['snp_%d' % i for i in range(len(snps))]
        else:
            print(snps)
            raise Exception('Invalid SNP table -- expected to find between 2 and 5 columns.')

        #fix data types
        snps = snps.astype(
            {
                'id': str,
                'chr': str,
                'pos': int,
            },
            errors = 'raise'
        )

        # #chr should always start with chr
        # if reference_type == 'genome':
        #     snps.loc[~snps['chr'].str.startswith('chr'), 'chr'] = ['chr' + c for c in snps.loc[~snps['chr'].str.startswith('chr'), 'chr']]

        #do we have genotypes for a snp?
        snps['snp_has_genotypes'] = False
        if 'snp_ref' in snps and 'snp_alt' in snps:
            snps.loc[snps['snp_ref'].notnull() & snps['snp_alt'].notnull(), 'snp_has_genotypes'] = True

        #stop if we have duplicate chr/pos
        if snps.duplicated(['chr', 'pos']).any():
            print(snps[snps.duplicated(['chr', 'pos'], keep=False)])
            raise Exception('Invalid SNPs table -- found duplicated chromosome/position!')

        #set the index to the snp id (stops if duplicate IDs)
        snps.set_index('id', drop = False, inplace = True, verify_integrity = True)

        log.info('Found SNP info for %d SNPs -- shape = %s', len(snps), str(snps.shape))
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = True).with_traceback(sys.exc_info()[2])

    return snps



#make sure we can import from package directory
import sys, os
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__)))) 

#use this to load parse_reads_cy properly
import pyximport; pyximport.install()

import unittest
import string, math
import amplimap.common
import amplimap.reader
from amplimap.parse_reads import *

class ParseReads(unittest.TestCase):
    def make_qual(self, qs, base = 33):
        return ''.join([ chr(base + max(0, min(40, q))) for q in qs ])
    def make_qual_from_p(self, ps, base = 33):
        return self.make_qual([ math.floor(-10 * math.log10(p)) for p in ps ], base)

    def test_mismatch_count_with_max(self):
        #perfect match
        self.assertEqual(mismatch_count_with_max(a=b"X"*10, b=b"X"*10, max_mismatches=0), 0)
        self.assertEqual(mismatch_count_with_max(a=b"X"*10, b=b"X"*10, max_mismatches=10), 0)
        #one mismatch
        self.assertEqual(mismatch_count_with_max(a=b"X"*10, b=b"X"*9 + b"B"*1, max_mismatches=0), 1)
        self.assertEqual(mismatch_count_with_max(a=b"X"*10, b=b"X"*9 + b"B"*1, max_mismatches=1), 1)
        self.assertEqual(mismatch_count_with_max(a=b"X"*10, b=b"X"*9 + b"B"*1, max_mismatches=10), 1)
        #two mismatches
        self.assertEqual(mismatch_count_with_max(a=b"A"+b"X"*9, b=b"X"*9 + b"B"*1, max_mismatches=0), 1) #we stop after 1
        self.assertEqual(mismatch_count_with_max(a=b"A"+b"X"*9, b=b"X"*9 + b"B"*1, max_mismatches=1), 2)
        self.assertEqual(mismatch_count_with_max(a=b"A"+b"X"*9, b=b"X"*9 + b"B"*1, max_mismatches=2), 2)
        self.assertEqual(mismatch_count_with_max(a=b"A"+b"X"*9, b=b"X"*9 + b"B"*1, max_mismatches=10), 2)

        #dots
        self.assertEqual(mismatch_count_with_max(a=b"X" + b"."*9, b=b"X"*9 + b"B"*1, max_mismatches=10), 0)
        self.assertEqual(mismatch_count_with_max(a=b"A" + b"."*9, b=b"X"*9 + b"B"*1, max_mismatches=10), 1)
        self.assertEqual(mismatch_count_with_max(a=b"."*10, b=b"X"*9 + b"B"*1, max_mismatches=10), 0)
        self.assertEqual(mismatch_count_with_max(a=b"A"*10, b=b"X"*9 + b".", max_mismatches=10), 9)

        #different length
        self.assertEqual(mismatch_count_with_max(a=b"A"*100, b=b"A"*10, max_mismatches=0), 0)
        self.assertEqual(mismatch_count_with_max(a=b"A"*100, b=b"X"*9 + b"B"*1, max_mismatches=0), 1)
        self.assertEqual(mismatch_count_with_max(a=b"A"*10, b=b"A"*100, max_mismatches=0), 90)
        self.assertEqual(mismatch_count_with_max(a=b"A"*10, b=b"A"*100, max_mismatches=90), 90)
        self.assertEqual(mismatch_count_with_max(a=b"A"*10, b=b"A"*100, max_mismatches=100), 90)

        #special cases
        self.assertEqual(mismatch_count_with_max(a=b"", b=b"", max_mismatches=10), 0)
        self.assertEqual(mismatch_count_with_max(a=b"", b=b"ABC", max_mismatches=10), 3)

        #errors
        with self.assertRaises(TypeError):
            mismatch_count_with_max(a=b"ABC", b=None, max_mismatches=10)
        with self.assertRaises(TypeError):
            mismatch_count_with_max(a=None, b=b"ABC", max_mismatches=10)

    def test_quality_trim_read(self):
        seq = (string.ascii_uppercase * 4)[0:100]

        qual_all_good = self.make_qual_from_p([0.001] * 100)
        qual_all_bad = self.make_qual_from_p([0.5] * 100)

        qual_all_good_base64 = self.make_qual_from_p([0.001] * 100, base = 64)

        #all or nothing
        self.assertEqual(quality_trim_read(seq, qual_all_good), (100, seq, qual_all_good))
        self.assertEqual(quality_trim_read(seq, qual_all_bad), (0, '', ''))

        #different phred bases
        self.assertEqual(quality_trim_read(seq, qual_all_good_base64, phred_base = 64), (100, seq, qual_all_good_base64))
        with self.assertRaises(AssertionError):
            quality_trim_read(seq, qual_all_good_base64)
        with self.assertRaises(AssertionError):
            quality_trim_read(seq, qual_all_good, phred_base = 64)

        #example from http://resources.qiagenbioinformatics.com/manuals/clcgenomicsworkbench/650/Quality_trimming.html
        #see quality_trim_test_spreadsheet.xls for these numbers
        qual_example = self.make_qual([4, 8, 15, 30, 32, 23, 10, 31, 31, 20, 28, 29, 30, 10, 31, 31, 20, 28, 32, 31, 32, 33, 29, 30, 31, 32, 32, 31, 32, 33, 33, 32, 33, 34, 30, 31, 32, 33, 33, 32, 33, 34, 23, 7, 7, 12])
        seq_example = seq[:len(qual_example)]
        for limit_example, exp_start, exp_end in [
            (0.0005, 33, 41),
            (0.001, 4, 41),
            (0.05, 2, 42),
            (0.2, 1, 45),
            (0.4, 0, 45),
        ]:
            self.assertEqual(
                quality_trim_read(seq_example, qual_example, p_threshold = limit_example),
                (exp_end - exp_start + 1, seq_example[exp_start:exp_end+1], qual_example[exp_start:exp_end+1]))

    def test_make_extended_read_name(self):
        for read_id in ['', 'asdfdasf::', '::1::1::1']:
            for probe in ['', 'probe-1', '_probe']:
                for umi in ['', 'ACG', 'ACGTTTTTTT']:
                    if len(read_id) == 0 or len(probe) == 0:
                        with self.assertRaises(Exception):
                            amplimap.common.parse_extended_read_name(amplimap.common.make_extended_read_name(read_id, probe, umi))
                    else:
                        self.assertEqual(
                            (read_id, probe, umi),
                            amplimap.common.parse_extended_read_name(amplimap.common.make_extended_read_name(read_id, probe, umi))
                        )

    def test_make_trimmed_read(self):
        for do_find_second_arm in [False]: #should add this -- find_arm_sequence and find_arm_sequence_mismatches
            for trim_min_length in [None, 20, 200]:
                for trim_primers in [True, False]:
                    for umi in ["99999", "", "1111111111111"]:
                        target_len = 100
                        arm_len = 10
                        other_arm_len = 20
                        umi_len = len(umi)

                        read_seq_trimmed = (string.ascii_uppercase * 4)[0:target_len]
                        read_qual_trimmed = self.make_qual_from_p([0.001] * target_len)

                        read_seq_maybe_primers = read_seq_trimmed if trim_primers else "X"*arm_len + read_seq_trimmed + "Y"*other_arm_len
                        read_qual_maybe_primers = read_qual_trimmed if trim_primers else self.make_qual_from_p([0.1] * (arm_len)) + read_qual_trimmed + self.make_qual_from_p([0.2] * (other_arm_len))

                        read_seq = umi + "X"*arm_len + read_seq_trimmed + "Y"*other_arm_len + umi
                        read_qual = self.make_qual_from_p([0.1] * (umi_len+arm_len)) + read_qual_trimmed + self.make_qual_from_p([0.2] * (other_arm_len+umi_len))
                        read_id = 'READ:ID'
                        read = ["%s adsfdasfsadfsadfsdaf" % read_id, read_seq, read_qual]

                        probe = "a-zA-Z0-9._-"

                        primer_trim_end = umi_len+arm_len+target_len+(0 if trim_primers else other_arm_len)
                        quality_trimmed_difference = 0

                        infos = 'pr:Z:%s\tum:Z:%s\tol:i:%d\tts:i:%d\tte:i:%d' % (probe, umi if umi_len > 0 else 'X', len(read_seq), umi_len+(arm_len if trim_primers else 0), primer_trim_end)
                        fastq_lines = (
                            '@%s\t%s\n' % (amplimap.common.make_extended_read_name(read_id, probe, umi), infos) +
                            '%s\n' % read_seq_maybe_primers +
                            '+\n' +
                            '%s\n' % read_qual_maybe_primers
                        )

                        if trim_min_length == 200:
                            expected = (None, primer_trim_end, False, 0, len(read_seq_maybe_primers))
                        else:
                            expected = (fastq_lines, primer_trim_end, True, quality_trimmed_difference, len(read_seq_maybe_primers))

                        #print(do_find_second_arm, trim_min_length, trim_primers, umi, '-> expected:')
                        #print(expected)

                        output = make_trimmed_read(
                            read_id,
                            read,
                            probe,
                            target_len,
                            umi,
                            umi_len,
                            arm_len,
                            other_arm_len,
                            trim_primers = trim_primers,
                            trim_min_length = trim_min_length
                        )

                        #print(do_find_second_arm, trim_min_length, trim_primers, umi, '-> actual:')
                        #print(output)

                        self.assertEqual(output, expected)

    def test_find_probe_from_reads(self):
        #arm sequences to simulate and search for
        probes_dict = {
            'first_primer_5to3': {
                'p1': 'A'*10, 'p2': 'A'*10, 'p3': 'B'*15, 'p4': 'C'*20, 'p5': 'Y'*10, 'p6': 'Z'*20
            },
            #in read orientation (will be reversed once to make fragment seq and back again to make read seq)
            'second_primer_5to3': {
                'p1': 'm'*15, 'p2': 'n'*5, 'p3': 'o'*15, 'p4': 'p'*1, 'p5': 'q'*15, 'p6': 'r'*10
            }
        }

        #convert input primers into expected format
        first_arms_probes, first_arms_seq = zip(*probes_dict['first_primer_5to3'].items()) #split into lists

        #convert to utf-8
        first_arms_seq = tuple([seq.encode('utf-8') for seq in first_arms_seq])
        first_arms_dict = dict([(key, val.encode('utf-8')) for key, val in probes_dict['first_primer_5to3'].items()])
        second_arms_dict = dict([(key, val.encode('utf-8')) for key, val in probes_dict['second_primer_5to3'].items()])

        #length of the target region (between arms)
        for target_len in [100, 10]:
            #first arm
            for probe_first_arm in first_arms_dict.keys():
                #second arm
                for probe_second_arm in second_arms_dict.keys():
                    #length of first umi
                    for umi_one in [0, 5]:
                        #length of second umi
                        for umi_two in [0, 5]:
                            #number of mismatches to allow
                            for mismatches in [0, 2, 5]:
                                #where to add mismatches
                                for add_mismatch_to in ['first', 'both']:
                                    #how many mismatches to add
                                    for add_mismatches in [0, 1, 2, 3]:
                                        #read length to simulate
                                        for read_len in [10, 11, 75, 200]:
                                            #arms
                                            first_arm = first_arms_dict[probe_first_arm]
                                            second_arm = second_arms_dict[probe_second_arm]

                                            #target sequence
                                            fragment_seq_trimmed = (string.ascii_uppercase * 4)[0:target_len].encode('utf8')

                                            #add mismatches to arms
                                            for mismatch_ix in range(add_mismatches):
                                                if mismatch_ix < len(first_arm):
                                                    first_arm = first_arm[:mismatch_ix] + b'#' + first_arm[(mismatch_ix+1):]
                                                if add_mismatch_to == 'both' and mismatch_ix < len(second_arm):
                                                    second_arm = second_arm[:mismatch_ix] + b'#' + second_arm[(mismatch_ix+1):]

                                            #build whole region seq
                                            fragment_seq_with_primers = first_arm + fragment_seq_trimmed + second_arm[::-1] #note reverse of second arm
                                            fragment_seq_with_primers_umis = b'1' * umi_one + fragment_seq_with_primers + b'2' * umi_two

                                            #simulate reads from this sequence
                                            first_read_str = fragment_seq_with_primers_umis[:read_len:1]
                                            second_read_str = fragment_seq_with_primers_umis[:-(read_len+1):-1] #:-1 reverses, but need to go to -(len+1) then

                                            expected = []
                                            #if first arm is 0 then second arm len can actually be just
                                            if read_len >= max(len(first_arm)+umi_one, len(second_arm)+umi_two):
                                                first_second_hybrid = (probe_first_arm == 'p1' and probe_second_arm == 'p2')
                                                second_first_hybrid = (probe_first_arm == 'p2' and probe_second_arm == 'p1')
                                                if (probe_first_arm == probe_second_arm) or first_second_hybrid or second_first_hybrid:
                                                    if add_mismatches <= mismatches:
                                                        expected.append([probe_first_arm if not first_second_hybrid and not second_first_hybrid else probe_second_arm,
                                                            add_mismatches, min(add_mismatches, len(second_arm)) if add_mismatch_to == 'both' else 0])

                                            #in this case we can still find the second arm pair
                                            #second probes matches first with five mismatches, since first arm is same and second arm just has 5 bp
                                            if probe_first_arm in ['p1', 'p2'] and probe_second_arm != 'p2':
                                                if read_len >= max(len(first_arm)+umi_one, len(second_arms_dict['p2'])+umi_two):
                                                    if mismatches >= 5 and add_mismatches <= mismatches:
                                                        expected.append(['p2', min(add_mismatches, len(first_arm)), 5])

                                            if probe_first_arm in ['p4'] and probe_second_arm != 'p4':
                                                if read_len >= max(len(first_arm)+umi_one, len(second_arms_dict['p4'])+umi_two):
                                                    if mismatches >= 1 and add_mismatches <= mismatches:
                                                        expected.append(['p4', min(add_mismatches, len(first_arm)), 1])


                                            output = find_probe_from_reads(
                                                first_arms_probes,
                                                first_arms_seq,
                                                second_arms_dict,
                                                first_read_str,
                                                second_read_str,
                                                umi_one = umi_one,
                                                umi_two = umi_two,
                                                mismatches = mismatches
                                            )

                                            try:
                                                self.assertEqual(
                                                    sorted(output, key=operator.itemgetter(0)),
                                                    sorted(expected, key=operator.itemgetter(0))
                                                )
                                            except AssertionError:
                                                #rerun with debug
                                                find_probe_from_reads(
                                                    first_arms_probes,
                                                    first_arms_seq,
                                                    second_arms_dict,
                                                    first_read_str,
                                                    second_read_str,
                                                    umi_one = umi_one,
                                                    umi_two = umi_two,
                                                    mismatches = mismatches,
                                                    debug = True
                                                )

                                                print('RL=', read_len, 'arms=', probe_first_arm, probe_second_arm, 'max mm =', mismatches, 'add mm =', add_mismatches, 'to', add_mismatch_to, '->')
                                                print(first_arm)
                                                print(second_arm)
                                                print('len normal=', max(len(first_arm), len(second_arm)) + max(umi_one, umi_two))
                                                print('len hybrid=', max(len(first_arm), len(second_arms_dict['p2'])) + max(umi_one, umi_two))
                                                print('total=', fragment_seq_with_primers_umis)
                                                print('1st=', first_read_str)
                                                print('2nd=', second_read_str)
                                                print('exp:')
                                                print(expected)
                                                print('obs:')
                                                print(output)
                                                raise


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
"""
This module contains methods for processing and aggregating coverage files generated by ``bedtools``.
"""

import pandas as pd
import numpy as np
import re
import os

from .reader import read_sample_info

cov_cols = ['Target', 'min_coverage', 'sum_coverage', 'basepairs', 'cov_per_bp', 'fraction_zero_coverage', 'fraction_10x_coverage', 'fraction_30x_coverage']
cov_cols_dtypes = dict(zip(cov_cols, [str, int, int, int, float, float]))

def fraction_zero_coverage(coverage):
    """Calculate fraction of bases with coverage 0."""
    return 1.0 * (coverage == 0).sum() / len(coverage)

def fraction_10x_coverage(coverage):
    """Calculate fraction of bases with coverage 10 or more."""
    return 1.0 * (coverage >= 10).sum() / len(coverage)

def fraction_30x_coverage(coverage):
    """Calculate fraction of bases with coverage 30 or more."""
    return 1.0 * (coverage >= 30).sum() / len(coverage)

def process_file(input: str, output: str):
    """Read raw bedtools coverage file, calculate summary statistics and output them as CSV file.

    Args:
        input: path to a bedtools coverage file
        output: path to the summary CSV file
    """

    #read bedtools output
    depth = pd.read_csv(input, sep='\t', names = ['chr', 'start_0', 'end', 'id', 'score', 'strand', 'position', 'coverage'], low_memory=False)

    #summarize
    summary = depth.groupby('id').aggregate({'coverage': [np.min, np.sum, len, np.mean, fraction_zero_coverage, fraction_10x_coverage, fraction_30x_coverage]})

    #make id index into normal column, then reset column names
    summary.reset_index(level=0, inplace=True)
    summary.columns = cov_cols

    #write file
    summary.to_csv(output, index = False)

def aggregate(input, output):
    """Read coverage summary files and create aggregate files.

    Args:
        input: dict containing 'csvs', the list of csvs fils to aggregate, and optionally 'sample_info', a table with additional sample annotation
        output: dict containing paths for output files: merged, min_coverage, cov_per_bp, fraction_zero_coverage
    """
    #load sample information table
    sample_info = None
    if 'sample_info' in input and len(input['sample_info']) > 0:
        sample_info = read_sample_info(input['sample_info'][0])

    merged = None
    for file in input['csvs']:
        sname = os.path.basename(file)            
        sname = re.sub(r'\.coverage\.csv$', '', sname)

        print('Reading', file, 'for', sname, '...')
        df = pd.read_csv(file,
            index_col = False,
            dtype = cov_cols_dtypes)
        df['Sample'] = sname
        print(sname, 'coverage data shape:', str(df.shape))

        if merged is None:
            merged = df
        else:
            merged = merged.append(df, ignore_index = True)

    assert merged is not None, \
        '\n\nABORTED: Did not find any coverage data!\n\n'
            
    print('Merged data shape:', str(merged.shape))
    print(merged.head())

    print('Duplicated:')
    print(merged[merged.duplicated(['Target', 'Sample'], keep=False)])

    if sample_info is not None:
        merged = merged.join(sample_info, on = ['Sample', 'Target'], how = 'left')

    #make matrices
    for column in ['min_coverage', 'cov_per_bp', 'fraction_zero_coverage']:
        pivoted = merged.pivot(index='Target', columns='Sample', values=column)
        print('Made pivot table for', column, ' with shape', str(pivoted.shape))
        pivoted.to_csv(output[column])
        print(output[column])

    #output full merged data set
    merged.to_csv(output['merged'], index = False)


# -*- coding: utf-8 -*-
"""
This module contains methods for generating pileups (per-base allele counts) using pysam/samtools.
"""

#python 3 compat
#http://python-future.org/compatible_idioms.html
from __future__ import print_function

import sys
import os
import re

import argparse

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
sh = logging.StreamHandler()
sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
log.addHandler(sh)

import time

#use pandas to read the CSV file and write output files
import pandas as pd
#for defaultdict + sorting
import collections, itertools
import operator

#for pileup
import pysam
#for getting ref base
import pyfaidx
#for debugging umi distance
import distance

#functions for reading input files
from .reader import read_snps_txt, read_new_probe_design, read_targets
from .common import parse_extended_read_name, find_umi_groups

#need these in various places
call_types_bases = ['A', 'C', 'G', 'T']
call_types = call_types_bases + ['INS', 'DEL', 'SKIP']

def get_al_mate_starts(al: pysam.AlignedSegment):
    """Get set of mate starts for AlignedSegment, always giving read1 first and read2 second"""
    return (al.reference_start, al.next_reference_start) if al.is_read1 else (al.next_reference_start, al.reference_start)

def get_pileup_row(chrom, pos_0, raw_coverage = 0, target_id = None, target_type = None, ref = None, validate_probe_targets = False):
    """Get ordered dict of columns for pileup table."""
    row = collections.OrderedDict()
    row['chr'] = chrom
    row['pos'] = pos_0 + 1
    if target_id is not None:
        row['target_id'] = target_id
    if target_type is not None:
        row['target_type'] = target_type
    if ref:
        row['ref'] = ref[chrom][pos_0:pos_0+1]
        row['alts'] = ''
    row['raw_coverage'] = raw_coverage

    row['umi_groups'] = 0
    row['unique_raw_umis'] = 0
    row['unique_mate_starts'] = 0
    row['group_below_min'] = 0
    row['group_no_majority'] = 0

    row['group_no_call'] = 0
    row['number_called'] = 0
    row['number_called_hq'] = 0

    row['filtered_umi_n'] = 0
    row['filtered_low_quality'] = 0
    row['filtered_flagged'] = 0
    row['filtered_softclipped'] = 0
    if validate_probe_targets:
        row['outside_probe_target'] = 0
    row['probes'] = ''

    row['overlapping_mates'] = 0

    #row['count_indel_pos'] = 0
    #row['count_indel_neg'] = 0
    for call in call_types:
        row['count_%s' % call] = 0
    for call in call_types:
        row['count_hq_%s' % call] = 0

    return row

class PileupRowFilterException(Exception):
    """Raised when row fails a pileup filter, with filter_column being the column to count this in."""
    def __init__(self, filter_column, skip_read_pair = True):
        self.filter_column = filter_column
        self.skip_read_pair = skip_read_pair

class PileupGroupFilterException(Exception):
    """Raised when UMI group fails a pileup filter, with filter_column being the column to count this in."""
    def __init__(self, filter_column):
        self.filter_column = filter_column

def process_pileup_read(
        pr,
        probes_dict,
        reference_name,
        reference_pos_0,
        read_metadata,
        min_mapq,
        ignore_groups,
        group_with_mate_positions,
        validate_probe_targets,
        filter_softclipped,
        no_probe_data,
        debug):
    """
    Process single pileup read.
    """
    #http://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment
    #http://pysam.readthedocs.io/en/latest/api.html#pysam.PileupColumn
    #http://pysam.readthedocs.io/en/latest/api.html#pysam.PileupRead

    #filter out low mapping quality
    if pr.alignment.mapping_quality < min_mapq:
        raise PileupRowFilterException('filtered_low_quality', skip_read_pair = False) #we don't want to skip the entire pair (for now, to keep compat with old branch)

    #ignore non-proper pair, qc fail, secondary alignments, ...
    #removed: (not pr.alignment.is_proper_pair) or
    #removed: (pr.alignment.is_secondary) or (for secondary alignments from bowtie2, which we would want to keep. hopefully should not matter otherwise since bwa never gives us secondary (?))
    if (pr.alignment.is_qcfail) or \
    (pr.alignment.mate_is_unmapped) or \
    (pr.alignment.is_supplementary):
        raise PileupRowFilterException('filtered_flagged')

    #filter out reads with softclipped bases
    if filter_softclipped:
        if 'S' in pr.alignment.cigarstring:
            raise PileupRowFilterException('filtered_softclipped', skip_read_pair = False) #we don't want to skip the entire pair!

    #get read metadata
    read_name = pr.alignment.query_name
    mate_starts = get_al_mate_starts(pr.alignment) if group_with_mate_positions else None
    if ignore_groups:
        #we haven't pre-parsed the read yet, so let's do it now
        read_umi = None

        #parse info from read name (should be bowtie2 compatible)
        if no_probe_data:
            read_probe = None
        else:
            _, read_probe, _ = parse_extended_read_name(read_name)
    else:
        #get read metadata from before (will raise if we didn't see this before, but we should always have)
        alignment_tuple = (read_name, mate_starts)
        read_name, read_probe, read_umi, _ = read_metadata[alignment_tuple]

    #filter out reads where the original UMI contains an N
    if not ignore_groups and read_umi is not None:
        if b'N' in read_umi:
            raise PileupRowFilterException('filtered_umi_n')

    if validate_probe_targets and read_probe is not None:
        if reference_pos_0 < probes_dict['target_start_0'][read_probe] \
        or reference_pos_0 >= probes_dict['target_end'][read_probe] \
        or reference_name != probes_dict['chr'][read_probe]:
            raise PileupRowFilterException('outside_probe_target', skip_read_pair = False)

    #figure out call and phred qual
    try:
        my_call = None
        my_phred = None
        if pr.indel > 0: #this peaks ahead!
            my_call = 'INS' #note: this replaces the basecall at the base preceding the INS!
            my_phred = 40
        elif pr.is_del: #pr.indel < 0 peaks ahead, so we really just want to use pr.is_del here!
            my_call = 'DEL'
            my_phred = 40 #may want to use query_position_or_next here to get qual of next base?
        elif pr.is_refskip:
            my_call = 'SKIP'
            my_phred = 40
        else:
            pos = pr.query_position
            query_sequence = pr.alignment.query_sequence

            if pos is None:
                print(pr)
                print(pr.indel)
                print(pr.is_del)
                raise Exception('query_position is None?!')
            elif pos >= len(query_sequence):
                raise Exception('pos = %d but len = %d' % (pos, len(query_sequence)))
            elif pos < 0:
                raise Exception('query position < 0: %d' % pos)

            my_call = query_sequence[pos]
            my_phred = pr.alignment.query_qualities[pos]
    except:
        print(reference_name, reference_pos_0)
        my_call = 'ERROR'
        raise

    return my_call, my_phred, read_name, read_probe, read_umi, mate_starts

def record_read_in_group(read_calls, my_call, my_phred, my_umi, read_name):
    """
    Add call data from read to read_calls dictionary of name -> call.
    """
    my_call_data = [my_call, my_phred, my_umi] #, my_start

    #do we already have a call from this read pair?
    if read_name in read_calls:
        assert my_umi == read_calls[read_name][2] #should always match!

        # #always take lowest start (or overwrite if None, although in that case we're probably None too)
        # if read_calls[read_name][3] is None or my_call_data[3] > read_calls[read_name][3]:
        #     my_call_data[3] = read_calls[read_name][3]

        #resolve overlap by taking higher-quality call
        if read_calls[read_name][1] < my_phred:
            read_calls[read_name] = my_call_data
        #handle case where we have the same quality for two different calls (get suspicious)
        elif read_calls[read_name][1] == my_phred and read_calls[read_name][0] != my_call:
            read_calls[read_name][0] = 'N'

        return True
    else:
        read_calls[read_name] = my_call_data
        return False

def get_group_consensus(group_calls, min_consensus_count = 1, min_consensus_fraction = 0.51, ignore_groups = False, debug = False):
    """
    Calculate consensus call, count and phred for UMI group.
    """
    group_call_counts = collections.defaultdict(int)
    group_calls_maxphred = collections.defaultdict(int)

    #count how often we saw each call
    for (my_call, my_phred, _) in group_calls:
        group_call_counts[my_call] += 1
        group_calls_maxphred[my_call] = max(group_calls_maxphred[my_call], my_phred)

    #find call with highest count
    #note: as long as min_consensus_fraction is >0.5 we should never end up with a duplicate call
    #in the end, since if we have a duplicate its fraction has to be <= 50%
    group_consensus_call, group_consensus_count = None, 0
    for my_call, my_count in group_call_counts.items():
        if my_count > group_consensus_count:
            group_consensus_call = my_call
            group_consensus_count = my_count
    group_consensus_phred = group_calls_maxphred[group_consensus_call]

    if debug:
        print('Group', '( len =', len(group_calls), ')', ' calls:',
            ', '.join(['%dx %s (q%d)' % (group_call_counts[k], k, group_calls_maxphred[k]) for k in group_call_counts.keys()]))

    #filter out non-consensus calls by raising a PileupGroupFilterException
    if not ignore_groups and (group_consensus_count < min_consensus_count):
        if debug:
            print('-> Below min count -- ', group_consensus_call, ' (', group_consensus_count, 'x ), max Q =', group_consensus_phred)
        raise PileupGroupFilterException('group_below_min')
    elif not ignore_groups and (group_consensus_count < len(group_calls) * min_consensus_fraction):
        if debug:
            print('-> No majority call -- ', group_consensus_call, ' (', group_consensus_count, 'x ), max Q =', group_consensus_phred)
        raise PileupGroupFilterException('group_no_majority')
    elif group_consensus_call == 'N':
        if debug:
            print('-> No base call (N) -- ', group_consensus_call, ' (', group_consensus_count, 'x ), max Q =', group_consensus_phred)
        raise PileupGroupFilterException('group_no_call')
    else:
        if debug:
            print('-> Consensus: ', group_consensus_call, ' (', group_consensus_count, 'x ), max Q =', group_consensus_phred)

    return group_consensus_call, group_consensus_count, group_consensus_phred

def process_pileup_row(
        row,
        seen_probes,
        call_groups,
        snps_dict,
        ignore_groups,
        min_consensus_count,
        min_consensus_fraction,
        min_baseq,
        ref,
        debug = False):
    """
    Generate a pileup table row from processed read data and calculate some additional stats and annotation.
    """
    #make a list of probes
    row['probes'] = ';'.join(sorted(seen_probes))

    phreds = list()
    call_phreds = collections.defaultdict(list)

    #process dictionary of best calls per read pair
    nonref_call_groups = []
    nonref_call_supporting_reads = []
    row['umi_groups'] = len(call_groups)
    for (group_name, my_group) in call_groups.items():
        try:
            #find consensus for this group
            group_consensus_call, group_consensus_count, group_consensus_phred = get_group_consensus(
                my_group.values(),
                min_consensus_count = min_consensus_count,
                min_consensus_fraction = min_consensus_fraction,
                ignore_groups = ignore_groups,
                debug = debug
            )

            #record phreds
            phreds.append(group_consensus_phred)
            call_phreds[group_consensus_call].append(group_consensus_phred)

            #make lists of the groups supporting the nonref call, and their number of consensus reads
            if ref and group_consensus_call != row['ref']:
                if len(nonref_call_groups) < 5:
                    nonref_call_groups.append(str(group_name))
                if len(nonref_call_supporting_reads) < 25:
                    nonref_call_supporting_reads.append(group_consensus_count)

            #record the call
            row['number_called'] += 1
            row['count_%s' % group_consensus_call] += 1

            if group_consensus_phred >= min_baseq:
                row['number_called_hq'] += 1
                row['count_hq_%s' % group_consensus_call] += 1
        except PileupGroupFilterException as ex:
            #if the row is filtered, we raise a PileupRowFilterException with message being the corresponding filter column
            row[ex.filter_column] += 1

    #mean phred score for position
    if len(phreds) > 0:
        row['phred_mean'] = 1.0 * sum(phreds) / len(phreds)
    else:
        row['phred_mean'] = None

    #mean phred score per count
    for call in call_types_bases:
        if call in call_phreds:
            row['phred_%s' % call] = 1.0 * sum(call_phreds[call]) / len(call_phreds[call])
        else:
            row['phred_%s' % call] = None

    #extract call counts again
    call_counts = [row['count_hq_%s' % call] for call in call_types]

    #ref/alts count
    if ref:
        if row['ref'] in call_types:
            row['ref_hq_count'] = row['count_hq_%s' % row['ref']]
            row['nonref_hq_count'] = sum([row['count_hq_%s' % call] for call in call_types if call != row['ref']])
            row['alts'] = ';'.join([call_types[i] for i in range(len(call_types)) if call_types[i] != row['ref'] and call_counts[i] > 0])
        else:
            row['ref_hq_count'] = None
            row['nonref_hq_count'] = None
            row['alts'] = None

        #make a list of umi groups with the nonref call
        row['nonref_call_groups_5'] = ';'.join(nonref_call_groups)
        if row['nonref_hq_count'] is not None and row['nonref_hq_count'] > 5:
            row['nonref_call_groups_5'] += ';...'
        #make a sorted list of umi group converage
        row['nonref_call_supporting_reads_25'] = ';'.join([str(x) for x in sorted(nonref_call_supporting_reads)])
        if row['nonref_hq_count'] is not None and row['nonref_hq_count'] > 25:
            row['nonref_call_supporting_reads_25'] += ';...'

    #snp ref/alt count
    if snps_dict is not None and ('snp_has_genotypes' in snps_dict):
        if snps_dict['snp_has_genotypes'][row['target_id']]:
            row['snp_ref_hq_count'] = row['count_hq_%s' % snps_dict['snp_ref'][row['target_id']]]
            row['snp_alt_hq_count'] = row['count_hq_%s' % snps_dict['snp_alt'][row['target_id']]]
        else:
            row['snp_ref_hq_count'] = None
            row['snp_alt_hq_count'] = None

    #maj/second/nonmaj count
    row['maj_hq_count'] = max(call_counts)
    if row['maj_hq_count'] > 0:
        #if we have a maj count we need to have called something
        assert row['number_called_hq'] > 0

        row['maj_hq_percent'] = 100.0 * row['maj_hq_count'] / row['number_called_hq']
        row['maj_hq_calls'] = ';'.join([call_types[i] for i in range(len(call_types)) if call_counts[i] == row['maj_hq_count']])
        row['maj_hq_phreds'] = ';'.join([str(row['phred_%s' % call_types_bases[i]]) for i in range(len(call_types_bases)) if call_counts[i] == row['maj_hq_count']])

        row['non_maj_hq_count'] = sum([cc for cc in call_counts if cc != row['maj_hq_count']])
        #row['non_maj_hq_percent'] = 100.0 * row['non_maj_hq_count'] / row['number_called_hq']

        non_maj = [i for i in range(len(call_types)) if call_counts[i] > 0 and call_counts[i] != row['maj_hq_count']]
        non_maj_counts = [call_counts[i] for i in non_maj]
        # non_maj_phreds = list(itertools.chain.from_iterable(( call_phreds[call_types[i]] for i in non_maj )))
        row['non_maj_hq_calls'] = ';'.join([call_types[i] for i in non_maj])
        #row['non_maj_mean_phreds'] = ';'.join([str(row['phred_%s' % call_types[i]]) if 'phred_%s' % call_types[i] in row else 'NA' for i in non_maj])
        #row['non_maj_hq_phreds_mean'] = 1.0 * sum(non_maj_phreds) / len(non_maj_phreds) if len(non_maj_phreds) > 0 else None

        if len(non_maj_counts) > 0:
            row['second_hq_count'] = max(non_maj_counts)
            row['second_hq_percent'] = 100.0 * row['second_hq_count'] / row['number_called_hq']

            if row['second_hq_count'] > 0:
                second = [i for i in range(len(call_types)) if call_counts[i] == row['second_hq_count']]
                row['second_hq_calls'] = ';'.join([call_types[i] for i in second])
            else:
                row['second_hq_calls'] = None
        else:
            row['second_hq_count'] = None
            row['second_hq_percent'] = None
            row['second_hq_calls'] = None
    else:
        row['maj_hq_percent'] = None
        row['maj_hq_calls'] = None
        row['maj_hq_phreds'] = None
        row['non_maj_hq_count'] = None
        row['non_maj_hq_calls'] = None
        row['second_hq_count'] = None
        row['second_hq_percent'] = None
        row['second_hq_calls'] = None

    return row

def process_pileup_base(
    ref,
    probes_dict,
    targets_dict,
    snps_dict,
    region_index,
    pileup_base,
    min_consensus_count,
    min_consensus_fraction,
    min_mapq,
    min_baseq,
    ignore_groups,
    group_with_mate_positions,
    validate_probe_targets,
    filter_softclipped,
    no_probe_data,
    read_metadata,
    filtered_pair_qnames,
    umi_to_group,
    debug
):
    """
    Process a single basepair of pileup, generating a pileup table row.
    """
    #pysam is zero-based! http://pysam.readthedocs.io/en/latest/glossary.html#term-region
    row = get_pileup_row(pileup_base.reference_name, pileup_base.reference_pos,
        pileup_base.nsegments,
        targets_dict['id'][region_index] if targets_dict is not None and region_index < len(targets_dict['id']) else None,
        targets_dict['type'][region_index] if targets_dict is not None and region_index < len(targets_dict['id']) else None,
        ref,
        validate_probe_targets = validate_probe_targets)

    #loop over pileups, make a dict of the best-phred call per group per read id
    call_groups = collections.defaultdict(dict)
    seen_probes = set()
    seen_raw_umis = set()
    seen_mate_starts = set()
    for pr in pileup_base.pileups:
        #if we have already filtered this read just skip it immediately
        if pr.alignment.query_name in filtered_pair_qnames:
            row[filtered_pair_qnames[pr.alignment.query_name]] += 1
        else:
            try:
                read_call, read_phred, read_name, read_probe, read_umi, mate_starts = process_pileup_read(
                    pr,
                    probes_dict,
                    pileup_base.reference_name,
                    pileup_base.reference_pos,
                    read_metadata,
                    min_mapq = min_mapq,
                    ignore_groups = ignore_groups,
                    group_with_mate_positions = group_with_mate_positions,
                    validate_probe_targets = validate_probe_targets,
                    filter_softclipped = filter_softclipped,
                    no_probe_data = no_probe_data,
                    debug = debug)

                #record which probes we have seen
                if read_probe is not None and not read_probe in seen_probes:
                    seen_probes.add(read_probe)

                #record which mate starts we have seen
                seen_mate_starts.add(mate_starts)

                #figure out umi (needed for grouping)
                if not ignore_groups:
                    #determine group
                    my_group = umi_to_group[mate_starts][read_umi]
                    seen_raw_umis.add(read_umi)
                else:
                    #just group by read name (every read is its own group)
                    my_group = read_name

                #record read straight in umi group - should be fast
                is_overlap = record_read_in_group(call_groups[my_group], read_call, read_phred, read_umi, read_name)
                if is_overlap:
                    row['overlapping_mates'] += 1
            except PileupRowFilterException as ex:
                #if the row is filtered, we raise a PileupRowFilterException with message being the corresponding filter column
                row[ex.filter_column] += 1

                #this tells us whether we should always skip this read, or just skip this position of this read
                if ex.skip_read_pair:
                    filtered_pair_qnames[pr.alignment.query_name] = ex.filter_column

    row['unique_raw_umis'] = len(seen_raw_umis)
    row['unique_mate_starts'] = len(seen_mate_starts)

    row = process_pileup_row(
        row,
        seen_probes,
        call_groups,
        snps_dict = snps_dict,
        ignore_groups = ignore_groups,
        min_consensus_count = min_consensus_count,
        min_consensus_fraction = min_consensus_fraction,
        min_baseq = min_baseq,
        ref = ref,
        debug = debug
        )

    return row


def aggregate(folder, snps_file, ref, generate_calls):
    log.info('Aggregating tables in %s', folder)

    #load SNPs to add SNP info
    snps = None
    if snps_file is not None and len(snps_file) > 0:
        snps = read_snps_txt(snps_file)

    agg = None
    agg_long = None
    coverage_long = None
    coverage_agg = None

    files = [file for file in os.listdir(folder) if file.endswith('.pileup.csv')]
    log.info('Found %d files', len(files))
    assert len(files) > 0

    files.sort()
    ixfile = 0
    for file in files:
        ixfile += 1
        sample = file
        sample = sample.replace('.pileup.csv', '')
        sample = sample.replace('_L001_R1_001', '')
        sample = sample.replace('__MIP_TRIMMED_', '')

        try:
            df = pd.read_csv(os.path.join(folder, file), low_memory=False, #otherwise get type warning, probably mis-detection of T/F or chr
                dtype = {'chr': str, 'pos': int, 'ref': str, 'alts': str})
            log.info('%d/%d: Found %d pileup rows in %s -> %s', ixfile, len(files), len(df), file, sample)
            if len(df) > 0:
                dup = df.duplicated(subset = ['chr', 'pos'], keep = 'first')
                if dup.any():
                    log.error('Found %d duplicate chr/pos pairs in file, cancelling!', dup.sum())
                    raise Exception()

                #add sample name and add to long table
                df['sample'] = sample

                #set filter status to NA
                df['filter'] = '--'

                #columns to join on
                cols = None
                #columns we actually want to output in the end
                out_cols = ['probes'] #always get the probes
                #columns that contain the actual counts
                count_cols = None #columns to output, but not join on

                #add snp info
                if snps is not None:
                    df = snps.merge(df, on = ('chr', 'pos'), how = 'left')
                    df.loc[df['snp_ref'].isnull(), 'snp_ref'] = ''
                    df.loc[df['snp_alt'].isnull(), 'snp_alt'] = ''

                    #use the snp columns as our baseline
                    cols = snps.columns.tolist()
                    count_cols = ['snp_ref_hq_count', 'snp_alt_hq_count']

                    #add filter stats
                    df['filter'] = ''

                    if ref:
                        #compare ref/observed alts to snp alleles (empty snp_ref/alt = pass)
                        df.loc[(df['snp_ref'] != '') & (df['ref'] != df['snp_ref']), 'filter'] += 'REF_mismatch;'
                        df.loc[(df['snp_alt'] != '') & (~df['alts'].isnull()) \
                            & (df['alts'] != df['snp_alt']), 'filter'] += 'unexpected_alt;'
                    else:
                        #compare observed alleles to snp alleles
                        df['filter'] = ['.' if pd.Series([row.maj_hq_calls] + row.second_hq_calls.split(';')).isin(
                            ['', row.snp_ref, row.snp_alt]).all() else 'unknown_allele' for row in df.itertuples()]

                        #warn if multiple major/second
                        df.loc[df['maj_hq_calls'].str.contains(';'), 'filter'] += 'multiple_major;'
                        df.loc[df['second_hq_calls'].str.contains(';'), 'filter'] += 'multiple_second;'

                    df.loc[df['filter'] == '', 'filter'] = '.' #set empty to dot
                else:
                    #columns to include in every row
                    cols = ['chr', 'pos']
                    #add target id, if we have it
                    if 'target_id' in df.columns:
                        cols.append('target_id')
                    if 'target_type' in df.columns:
                        cols.append('target_type')

                    if ref:
                        count_cols = ['ref_hq_count', 'nonref_hq_count']
                        cols.append('ref') #keep the ref column
                        cols.append('alts') #keep the alt column
                    else:
                        df.loc[df['maj_hq_calls'].isnull(), 'maj_hq_calls'] = ''
                        df.loc[df['second_hq_calls'].isnull(), 'second_hq_calls'] = ''

                        count_cols = ['maj_hq_count', 'second_hq_count']

                #also get the total number of called umis
                count_cols.append('number_called_hq')

                #make sure these are properly zero-filled
                for col in count_cols:
                    df[col] = df[col].fillna(0).astype(int)

                #add counts to output
                out_cols += count_cols

                #calculate fractions
                df[count_cols[0]+'_fraction'] = 1.0 * df[count_cols[0]] / df['number_called_hq']
                df[count_cols[1]+'_fraction'] = 1.0 * df[count_cols[1]] / df['number_called_hq']

                #either make this the first one or add to existing
                if agg_long is None:
                    agg_long = df
                else:
                    agg_long = pd.concat([agg_long, df], ignore_index=True, copy=True)

                #subset to merge_cols + output cols + fractions for wide table
                df = df.loc[:, cols + out_cols + [count_cols[0]+'_fraction', count_cols[1]+'_fraction', 'filter']]

                #drop alts for wide table
                if 'alts' in df.columns:
                    df.drop('alts', axis=1, inplace=True)

                #rename columns
                df.rename(columns = {'filter': sample+'_filter',
                    'maj_hq_count': sample + '_maj',
                    'second_hq_count': sample + '_sec',
                    'maj_hq_count_fraction': sample + '_fraction_maj',
                    'second_hq_count_fraction': sample + '_fraction_sec',
                    'ref_hq_count': sample + '_ref',
                    'nonref_hq_count': sample + '_alt',
                    'ref_hq_count_fraction': sample + '_fraction_ref',
                    'nonref_hq_count_fraction': sample + '_fraction_alt',
                    'snp_ref_hq_count': sample + '_ref',
                    'snp_alt_hq_count': sample + '_alt',
                    'snp_ref_hq_count_fraction': sample + '_fraction_ref',
                    'snp_alt_hq_count_fraction': sample + '_fraction_alt',
                    'number_called_hq': sample + '_total'},
                    inplace = True)

                if agg is None:
                    agg = df
                    agg['total_count'] = agg[sample + '_total'].fillna(0)
                else:
                    agg = agg.merge(df, how='outer', on = [c for c in cols if c != 'alts'], suffixes = ('', '_2'))

                    #add up total count, drop extra column
                    agg['total_count'] = agg['total_count'].fillna(0) + agg[sample + '_total'].fillna(0)
                    #if no probe yet set probes to our probe, check probes, drop extra column
                    agg.loc[agg['probes'].isnull(), 'probes'] = agg.loc[agg['probes'].isnull(), 'probes_2']
                    agg.loc[~agg['probes_2'].isnull() & (agg['probes'] != agg['probes_2']), sample+'_filter'] = 'probes_mismatch'
                    agg.drop('probes_2', axis=1, inplace=True)

                #stats
                log.info('%d/%d: Pileup rows processed, agg shape now %s, agg long shape now %s', ixfile, len(files), str(agg.shape), str(agg_long.shape))
        except pd.errors.EmptyDataError:
            log.exception("No data found: %s", file)

        coverage_file = file.replace('.pileup.csv', '.targets.csv')
        if os.path.exists(os.path.join(folder, coverage_file)):
            try:
                coverage_df = pd.read_csv(os.path.join(folder, coverage_file), low_memory=False)
                log.info('%d/%d: Found %d coverage rows in %s -> %s', ixfile, len(files), len(coverage_df), coverage_file, sample)

                if len(coverage_df) > 0:
                    coverage_df = coverage_df[ coverage_df['type'] == 'target' ]
                    log.info('%d/%d: %d coverage rows after removing non-targets', ixfile, len(files), len(coverage_df))

                    coverage_df['sample'] = sample
                    coverage_long = pd.concat([coverage_long, coverage_df], ignore_index=True, copy=True)

                    #remove some default columns that we don't need
                    coverage_df = coverage_df[ [c for c in coverage_df.columns if not c in ['chr', 'start_0', 'end', 'length', 'type']] ]
                    coverage_df.columns = [ c if c == 'id' else sample + '_' + c for c in coverage_df.columns ]

                    if coverage_agg is None:
                        coverage_agg = coverage_df
                    else:
                        coverage_agg = coverage_agg.merge(coverage_df, how='outer', on = 'id')
            except pd.errors.EmptyDataError:
                log.exception("No data found: %s", coverage_file)

    if agg is None:
        raise Exception('No data found (in %d files), aborting!' % len(files))

    log.info('Agg Wide: Found %d rows total, %d columns', len(agg), agg.shape[1])
    if snps is None:
        agg = agg[agg['total_count'] > 0]
        log.info('%d rows left after removing zero-counts', len(agg))

    log.info('Agg Long: Found %d rows total, %d columns', len(agg_long), agg_long.shape[1])

    if coverage_agg is not None:
        log.info('Cov Wide: Found %d coverage rows total, %d columns', len(coverage_agg), coverage_agg.shape[1])

    if coverage_long is not None:
        log.info('Cov Long: Found %d coverage rows total, %d columns', len(coverage_long), coverage_long.shape[1])

    agg['chr_num'] = [ 31 if chr[3:] == 'X' else 32 if chr[3:] == 'Y' else 33 if chr[3:] == 'M' else int(chr[3:]) if chr[3:].isdigit() else 99 for chr in agg['chr'] ]
    agg.sort_values(['chr_num', 'chr', 'pos'], inplace=True)
    agg.drop('chr_num', axis=1, inplace=True)

    outname = os.path.join(folder, '%spileups_wide.csv' % ('' if snps is None else 'target_snps_'))
    agg.to_csv(outname, index=False)
    print(outname)

    #figure out the baseline columns
    agg_columns = ['sample']
    if snps is not None:
        agg_columns += snps.columns.tolist()
    else:
        #columns to include in every row
        agg_columns += ['chr', 'pos']

        #add target id, if we have it
        if 'target_id' in df.columns:
            agg_columns.append('target_id')
        if 'target_type' in df.columns:
            agg_columns.append('target_type')

        if ref:
            agg_columns.append('ref') #keep the ref column
            agg_columns.append('alts') #keep the alt column
    agg_columns.append('probes')

    #make two separate lists of columns for the two output files
    agg_long_columns = agg_columns.copy()
    agg_long_detailed_columns = agg_columns.copy()

    #figure out columns for long output
    if snps is not None:
        agg_long_count_cols = ['snp_ref_hq_count', 'snp_alt_hq_count']
    elif ref:
        agg_long_count_cols = ['ref_hq_count', 'nonref_hq_count']
    else:
        agg_long_count_cols = ['maj_hq_count', 'second_hq_count']
    agg_long_columns += agg_long_count_cols
    agg_long_columns.append('number_called_hq')
    agg_long_columns += ['%s_fraction' % count_col for count_col in agg_long_count_cols]
    agg_long_columns.append('filter')

    #figure out columns for detailed output
    agg_long_detailed_columns += ['count_hq_%s' % call for call in call_types]
    agg_long_detailed_columns.append('number_called_hq')
    agg_long_detailed_columns += ['%s_fraction' % count_col for count_col in agg_long_count_cols]
    agg_long_detailed_columns.append('filter')

    #output long table
    outname = os.path.join(folder, '%spileups_long.csv' % ('' if snps is None else 'target_snps_'))
    agg_long[agg_long_columns].to_csv(outname, index=False)
    print(outname)

    #output nucleotide count table
    outname = os.path.join(folder, '%spileups_long_detailed.csv' % ('' if snps is None else 'target_snps_'))
    agg_long[agg_long_detailed_columns].to_csv(outname, index=False)
    print(outname)

    #output coverage, but not if we are running SNPs
    if coverage_agg is not None and snps is None:
        outname = os.path.join(folder, 'target_coverage.csv')
        coverage_agg.to_csv(outname, index=False)
        print(outname)

    #output long coverage
    if coverage_long is not None and snps is None:
        outname = os.path.join(folder, 'target_coverage_long.csv')
        coverage_long.to_csv(outname, index=False)
        print(outname)

    if generate_calls:
        #ideas:
        # http://archive.broadinstitute.org/cancer/cga/mutect
        # http://www.nature.com/nbt/journal/v31/n3/abs/nbt.2514.html
        # http://dkoboldt.github.io/varscan/germline-calling.html

        if snps is not None:
            sys.exit('Cannot call variants when analysing known SNPs')
        if not ref:
            sys.exit('Cannot call variants without reference genome')

        def call_variants(pileup_rows):
            #use: 'ref_hq_count', 'nonref_hq_count, alts
            pass

        grouped_pileups = agg_long.groupby(['chr', 'pos', 'ref'])
        variant_calls = grouped_pileups.aggregate(call_variants)

        outname = os.path.join(folder, 'variant_calls.csv')
        variant_calls.to_csv(outname, index=False)
        print(outname)

def process_file(input, output, probes_file, snps_file, targets_file, validate_probe_targets,
    fasta_file,
    min_mapq, min_baseq, ignore_groups,
    min_consensus_count,
    min_consensus_fraction = 0.51,
    group_with_mate_positions = False,
    filter_softclipped = True,
    ignore_duplicates = False,
    no_probe_data = False,
    umi_tag_name = None,
    reference_type = 'genome',
    subsample_reads = None,
    debug_umi_groups = False, debug_pos = None, debug = False
):
    if subsample_reads:
        log.info('Subsampling reads to at most %d per position!', subsample_reads)

    #load probes for target validation
    probes_dict = None
    if validate_probe_targets:
        assert not no_probe_data, 'you cannot validate probe targets with --no-probe-data'
        if probes_file is not None and len(probes_file) > 0:
            probes = read_new_probe_design(probes_file)
            probes_dict = probes.to_dict()

    #load SNPs
    snps = None
    snps_dict = None
    if snps_file is not None and len(snps_file) > 0:
        snps = read_snps_txt(snps_file, reference_type = reference_type)
        snps_dict = snps.to_dict()

    #load target regions
    targets = None
    if targets_file is not None and len(targets_file) > 0:
        assert snps is None, 'cannot pileup on target regions and SNPs at the same time!'
        targets = read_targets(targets_file, reference_type = reference_type, file_type = 'bed')
    elif snps is not None:
        #only pileup on SNPs. otherwise SNP in target regions have the wrong target id
        targets = pd.DataFrame({'chr': snps.chr, 'start_0': snps.pos - 1, 'end': snps.pos, 'id': snps.id, 'type': 'snp'})
        targets['start_0'] = targets['start_0'].astype(int)
        targets['end'] = targets['end'].astype(int)
        targets['length'] = targets['end'] - targets['start_0']
        targets.index = range(len(targets))

    #load reference genome
    ref = None
    if fasta_file is not None and len(fasta_file) > 0:
        ref = pyfaidx.Fasta(fasta_file, read_ahead=1000, as_raw=True, sequence_always_upper=True, rebuild=False)
        log.info('Loaded reference genome fasta from %s', fasta_file)

    #figure out actual ends
    #TODO: this is actually no longer in use, since reference_type is forced to 'genome' now.
    if reference_type == 'transcriptome' and targets is not None:
        if ref is None:
            raise Exception('Error: need reference fasta file to do pileup on transcriptome!')

        try:
            targets['end'] = [ len(ref[row.chr]) for row in targets.itertuples() ]
        except KeyError:
            raise
        targets['length'] = targets['end'] - targets['start_0']
        log.info('Fixed target end and lengths')
        print(targets)

    #figure out pileup regions
    pileup_regions = [None]
    if targets is not None:
        #pysam start is 0-based, so we can keep this as is
        pileup_regions = [ (row.chr, row.start_0, row.end, row.id) for row in targets.itertuples() ]

    #no make targets dict (which may actually be SNPs)
    targets_dict = None
    if targets is not None:
        targets_dict = targets.to_dict()

    if len(pileup_regions) > 0 and pileup_regions[0] is not None:
        log.info('Got %d pileup regions: %s ...', len(pileup_regions), str(pileup_regions[0:3]))
    else:
        log.info('Got no specific pileup regions - piling up everywhere!')

    #prepare coverage columns
    if not targets is None:
        targets['coverage_min'] = None
        targets['coverage_avg'] = None
        targets['coverage_zero_percentage'] = None
        targets['called_hq_min'] = None
        targets['called_hq_avg'] = None
        targets['called_hq_zero_percentage'] = None

    log.info('Processing file: %s', input)
    if not os.path.isfile(input):
        sys.exit('input does not exist: %s' % input)

    with pysam.AlignmentFile(input, "rb") as samfile:
        t_shown = t_start = time.time()

        rows = []
        for region_index, region in enumerate(pileup_regions):
            region_length = None
            max_depth = 1e7 if not subsample_reads else subsample_reads

            #our region
            assert region is not None
            region_id = region[3]
            region_length = region[2] - region[1]
            region_coverage = [0] * region_length
            region_coverage_hq = [0] * region_length

            #first figure out umi groups
            umi_to_group = dict()
            read_metadata = dict()
            if no_probe_data and not umi_tag_name:
                log.info('No UMI info, not grouping reads!')
            elif ignore_groups:
                log.info('Ignore groups is set, not grouping reads!')
            else:
                log.info('Grouping reads in region #%d: %s (length = %d)', region_index, region, region_length)
                iter_grouping = samfile.fetch(region[0], region[1], region[2])

                if debug_umi_groups:
                    mate_start_counts = collections.defaultdict(collections.Counter)
                    read_alignment_counter = collections.Counter()

                umi_counts = collections.defaultdict(collections.Counter)
                n_alignments = 0
                for al in iter_grouping:
                    n_alignments += 1
                    raw_read_name = al.query_name

                    #are we dealing with reads that don't have probe data attached?
                    if no_probe_data:
                        original_read_name = raw_read_name
                        read_probe = None
                        read_umi = None

                        #get UMI from tag then!
                        if not umi_tag_name is None:
                            try:
                                read_umi = al.get_tag(umi_tag_name)
                            except KeyError:
                                raise Exception('Encountered alignment with missing UMI tag: "%s", was looking for "%s"' % (raw_read_name, umi_tag_name))
                    else:
                        #parse info from read name (should be bowtie2 compatible)
                        original_read_name, read_probe, read_umi = parse_extended_read_name(raw_read_name)

                    #count this
                    if debug_umi_groups:
                        read_alignment_counter[original_read_name] += 1

                    #make sure we have bytes for read_umi
                    assert read_umi is not None and len(read_umi) > 0
                    read_umi = read_umi.encode('utf-8')

                    #use starts of both mates to uniquely identify alignment
                    mate_starts = get_al_mate_starts(al) if group_with_mate_positions else None

                    #only do this once per read pair and start
                    alignment_tuple = (raw_read_name, mate_starts)
                    read_metadata_tuple = (original_read_name, read_probe, read_umi, [n_alignments])
                    if alignment_tuple in read_metadata:
                        #check metadata (except for n_alignments)
                        for ix in range(3):
                            assert read_metadata[alignment_tuple][ix] == read_metadata_tuple[ix]
                        #add read index
                        read_metadata[alignment_tuple][3].append(n_alignments)
                    else:
                        #record metadata
                        read_metadata[alignment_tuple] = read_metadata_tuple

                        #record umi (needed for grouping)
                        umi_counts[mate_starts][read_umi] += 1
                        if debug_umi_groups:
                            mate_start_counts[read_umi][mate_starts] += 1

                if debug_umi_groups:
                    aligned_reads_per_read_pair = collections.Counter()
                    for v in read_alignment_counter.values():
                        aligned_reads_per_read_pair[v] += 1
                    log.info('Seen alignments per read pair (expect 1/2): %s', str(aligned_reads_per_read_pair))

                    reads_per_alignment = collections.Counter()
                    for v in read_metadata.values():
                        assert len(v[3]) in [1, 2] #we expect to see either 1 or 2 reads for each read_name/mate_positions combo
                        reads_per_alignment[len(v[3])] += 1
                    log.info('Seen reads per alignment (expect 1/2): %s', str(reads_per_alignment))

                    starts_per_umi = collections.Counter()
                    for read_umi, mate_counts_for_umi in mate_start_counts.items():
                        starts_per_umi[len(mate_counts_for_umi)] += 1
                        if len(mate_counts_for_umi) > 1:
                            log.info('Seen %s with multiple mate starts: %s', str(read_umi), str(mate_counts_for_umi.keys()))
                    log.info('Seen mate positions per UMI (expect 1): %s', str(starts_per_umi))

                #free up memory
                log.info('%s: Processed UMIs from %d read pairs with %d alignments.', region_id, len(read_metadata), n_alignments)

                umi_id_offset = 0
                target_raw_umi_count = 0
                if debug_umi_groups:
                    umi_count_histogram = collections.Counter()
                    umi_count_histogram_umis = collections.defaultdict(list)
                #log.info('%s: Got %d different subgroups (starts?) for UMIs.', region_id, len(umi_counts))
                for mate_starts, umi_counts_for_start in umi_counts.items():
                    target_raw_umi_count += len(umi_counts_for_start)

                    if debug_umi_groups:
                        for read_umi, pairs_for_umi in umi_counts_for_start.items():
                            umi_count_histogram[pairs_for_umi] += 1
                            if pairs_for_umi > 5:
                                umi_count_histogram_umis[pairs_for_umi].append(read_umi)

                    #use umi_tools to group UMIs together
                    umi_to_group[mate_starts] = find_umi_groups(umi_counts_for_start, umi_id_offset)
                    umi_id_offset = max(umi_to_group[mate_starts].values()) + 1

                    #log.info('%s: Recorded groups for %d raw UMIs with start = %s, new offset: %d', region_id, len(umi_counts_for_start), str(mate_starts), umi_id_offset)
                log.info('%s: Grouped %d raw UMIs with %d different mate start combinations', region_id, target_raw_umi_count, len(umi_counts))

                if debug_umi_groups:
                    log.info('Raw UMI count hist:')
                    for count in sorted(umi_count_histogram.keys()):
                        log.info('  %d\t%d\t%s', count, umi_count_histogram[count], str(umi_count_histogram_umis[count][:5]) if count > 5 else '')

                #free up memory
                del umi_counts

            #then do actual pileup
            #note: do not filter at this stage since we will do this ourselves
            iter_pileup = samfile.pileup(region[0], region[1], region[2], max_depth = max_depth, stepper = 'nofilter')
            log.info('Piling up in region #%d: %s (length = %d)', region_index, region, region_length)

            seen_positions = set()
            filtered_pair_qnames = dict()
            for pileup_base in iter_pileup:
                #make sure we don't venture outside the regions, even if we have reads there
                if region is not None:
                    if pileup_base.reference_pos < region[1] or pileup_base.reference_pos >= region[2]:
                        #log.info('Skipping calls at 0-based position %d (outside of target region)', pileup_base.reference_pos)
                        continue

                #if we have a debug pos, ignore everything else!
                if debug_pos is not None:
                    chr_pos = debug_pos.split(':')
                    if pileup_base.reference_name != chr_pos[0] or str(pileup_base.reference_pos + 1) != chr_pos[1]:
                        continue

                #we need to set a max_depth above, but we can make it high
                if not subsample_reads:
                    assert pileup_base.nsegments < max_depth, 'maximum coverage %d reached, increase limit!' % max_depth

                #record row
                row = process_pileup_base(
                    ref,
                    probes_dict,
                    targets_dict,
                    snps_dict,
                    region_index,
                    pileup_base,
                    min_consensus_count,
                    min_consensus_fraction,
                    min_mapq,
                    min_baseq,
                    ignore_groups,
                    group_with_mate_positions,
                    validate_probe_targets,
                    filter_softclipped,
                    no_probe_data,
                    read_metadata,
                    filtered_pair_qnames,
                    umi_to_group,
                    debug
                )
                rows.append(row)
                seen_positions.add(pileup_base.reference_pos)

                #record coverage
                if region is not None:
                    coverage_index = pileup_base.reference_pos - region[1]
                    if coverage_index >= 0 and coverage_index < region_length:
                        region_coverage[coverage_index] = row['raw_coverage']
                        region_coverage_hq[coverage_index] = row['number_called_hq']

                t_now = time.time()
                if len(rows) == 1 or len(rows) == 10 or len(rows) % 10000 == 0 or (t_now - t_shown > 60):
                    t_shown = t_now
                    log.info("processed %d positions - %.f sec elapsed, %.4f sec/position, %.1f positions/sec",
                        len(rows),
                        t_now - t_start,
                        (t_now - t_start) / len(rows),
                        len(rows) / (t_now - t_start))
                    #log.info(str(row))

            if region is not None:
                #make sure we have pileup rows for every single nucleotide in the region
                #log.info('%s: Got %d pileup rows so far - filling in missing positions between %d and %d (0-based)', region_id, len(rows), region[1], region[2])
                for pos_0 in range(region[1], region[2]):
                    if not pos_0 in seen_positions:
                        #log.info('Row missing for: %d (0-based)', pos_0)
                        #TODO: could refactor this into just doing get/process in one go, now that we call this anywy
                        row = get_pileup_row(region[0], pos_0,
                            0,
                            targets_dict['id'][region_index] if targets_dict is not None and region_index < len(targets_dict['id']) else None,
                            targets_dict['type'][region_index] if targets_dict is not None and region_index < len(targets_dict['id']) else None,
                            ref,
                            validate_probe_targets = validate_probe_targets)

                        row = process_pileup_row(
                            row,
                            seen_probes = set(),
                            call_groups = dict(),
                            snps_dict = snps_dict,
                            ignore_groups = ignore_groups,
                            min_consensus_count = min_consensus_count,
                            min_consensus_fraction = min_consensus_fraction,
                            min_baseq = min_baseq,
                            ref = ref,
                            debug = debug
                            )
                        rows.append(row)
                log.info('%s: %d pileup rows', region_id, len(rows))

                #calculate overall coverage
                coverage_min = min(region_coverage)
                coverage_avg = 1.0 * sum(region_coverage) / region_length
                coverage_zeros = region_coverage.count(0)
                log.info('%s coverage (raw): min=%d, avg=%f, zeros=%d', region_id, coverage_min, coverage_avg, coverage_zeros)

                coverage_hq_min = min(region_coverage_hq)
                coverage_hq_avg = 1.0 * sum(region_coverage_hq) / region_length
                coverage_hq_zeros = region_coverage_hq.count(0)
                log.info('%s coverage (called + baseq > %d): min=%d, avg=%f, zeros=%d', region_id, min_baseq, coverage_hq_min, coverage_hq_avg, coverage_hq_zeros)

                if targets is not None and region_index < len(targets):
                    targets.loc[region_index, 'coverage_min'] = coverage_min
                    targets.loc[region_index, 'coverage_avg'] = coverage_avg
                    targets.loc[region_index, 'coverage_zero_percentage'] = 100.0 * coverage_zeros / region_length
                    targets.loc[region_index, 'called_hq_min'] = coverage_hq_min
                    targets.loc[region_index, 'called_hq_avg'] = coverage_hq_avg
                    targets.loc[region_index, 'called_hq_zero_percentage'] = 100.0 * coverage_hq_zeros / region_length
                else:
                    log.info('Not recording coverage for region #%d because it is not a target region (SNP?)', region_index)

        log.info('Pileups resulted in %d rows', len(rows))
        outname = output + '.pileup.csv'
        if len(rows) > 0:
            df = pd.DataFrame(rows)

            #remove duplicate rows
            dup = df.duplicated(subset = ['chr', 'pos'], keep = 'first')
            if dup.any():
                if ignore_duplicates:
                    log.warn('Dropping %d/%d duplicated rows', dup.sum(), len(df))
                    df = df[~dup]
                    log.info('Now have %d rows after dropping dupes', len(df))
                else:
                    print(df[dup, 'target_id'].value_counts())
                    sys.exit('Found duplicate rows, cancelling! Make sure none of the regions in targets.bed overlap.')
            else:
                log.info('No duplicates found')

            #sort rows (may have become unsorted due to targest)
            df['chr_num'] = [ 31 if chr[3:] == 'X' else 32 if chr[3:] == 'Y' else 33 if chr[3:] == 'M' else int(chr[3:]) if chr[3:].isdigit() else 99 for chr in df['chr'] ]
            df.sort_values(['chr_num', 'chr', 'pos'], inplace=True)
            df.drop('chr_num', axis=1, inplace=True)

            #for call in call_types:
            #    df['percent_%s' % call] = 100.0 * df['count_%s' % call] / df['number_called']
            #for call in call_types:
            #    df['percent_hq_%s' % call] = 100.0 * df['count_hq_%s' % call] / df['number_called_hq']

            df.to_csv(outname, index=False)
        else:
            log.warn('No data to output, creating empty file!')
            open(outname, 'a').close()

        print(outname)

        if snps is None:
            if targets is not None:
                outname = output + '.targets.csv'
                targets.to_csv(outname, index=False)
                print(outname)

def main():
    print('Called with arguments: "%s"' % '" "'.join(sys.argv))

    #parse the arguments, which will be available as properties of args (e.g. args.probe)
    parser = argparse.ArgumentParser()
    #specify parameters
    parser.add_argument("--aggregate", help="folder with processed CSV files to aggregate")

    parser.add_argument("--input", help="input bam file name")
    parser.add_argument("--output", help="output file prefix")
    parser.add_argument("--probes", help="CSV file with probes and their arm sequences")
    parser.add_argument("--targets", help="target regions bed file")
    parser.add_argument("--snps", help="target SNPs file")

    parser.add_argument("--ref", help="find reference base at each position (requires --fasta)", action='store_true')
    parser.add_argument("--fasta", help="reference genome fasta (needs faidx)", default='/home/crangen/zding/projects/common/ref/old/hg19_GRCh37/hg19.fa')

    #parser.add_argument("--generate-calls", help="generate variant calls", action='store_true') TODO

    parser.add_argument("--ignore-groups", help="ignore the UMI group entirely", action='store_true')
    parser.add_argument("--group-with-mate-positions", help="only group read pairs if the aligned start positions of both mates are the same", action='store_true')
    parser.add_argument("--min-consensus-count", help="minimum number of consistent read pairs supporting a UMI", default=2, type=int)
    parser.add_argument("--min-consensus-percentage", help="minimum fraction of reads supporting the consensus call in a UMI group", default=51, type=int)

    parser.add_argument("--min-mapq", help="minimum mapping quality (remove read pairs with either read having mapq < X)", default=20, type=int)
    parser.add_argument("--min-baseq", help="minimum base quality (ignore calls from groups below baseq < X at this position)", default=30, type=int)
    parser.add_argument("--ignore-duplicates", help="ignore duplicate positions caused by overlapping targets", action='store_true')
    parser.add_argument("--validate-probe-targets", help="filter base calls outside of the actual probe target regions (eg. if primers have not been trimmed already)", action='store_true')

    parser.add_argument("--no-filter-softclipped", help="do not filter reads that contain softclipped bases", action='store_true')
    parser.add_argument("--no-probe-data", help="do not try to read probe data", action='store_true')
    parser.add_argument("--umi-tag-name", help="name of the BAM tag containing UMIs")

    parser.add_argument("--debug", help="enable debug output", action='store_true')
    parser.add_argument("--debug-pos", help="only do pileup at this position (chr:pos)")
    parser.add_argument("--debug-umi-groups", help="debug umi groups", action='store_true')
    args = parser.parse_args()
    args.generate_calls = None #TODO

    if args.aggregate:
        aggregate(args.aggregate, args.snps, args.ref, args.generate_calls)
    else:
        assert not (args.ref and args.fasta is None)
        assert args.input is not None
        assert args.output is not None

        process_file(args.input, args.output,
            args.probes, args.snps, args.targets, args.validate_probe_targets,
            args.fasta if args.ref else None,
            args.min_mapq, args.min_baseq,
            ignore_groups = args.ignore_groups,
            group_with_mate_positions = args.group_with_mate_positions,
            min_consensus_count = args.min_consensus_count,
            min_consensus_fraction = args.min_consensus_percentage / 100.0,
            filter_softclipped = not args.no_filter_softclipped,
            ignore_duplicates = args.ignore_duplicates,
            no_probe_data = args.no_probe_data,
            umi_tag_name = args.umi_tag_name,
            debug_umi_groups = args.debug_umi_groups,
            debug_pos = args.debug_pos,
            debug = args.debug)

    return 0


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
if sys.version_info[0] < 3:
    sys.exit("Sorry, amplimap requires at least Python 3")

from setuptools import setup, find_packages, Extension

#load version number (can't just import the package since we might miss requirements)
filename = 'amplimap/version.py'
exec(compile(open(filename, "rb").read(), filename, 'exec')) #python2/3 compatible replacement for execfile()

#load long description from readme
import codecs
with codecs.open('README.rst', 'r', 'utf-8') as f:
    long_description = f.read()

setup(
    name = __title__,
    version = __version__,

    packages = find_packages(),
    #these files will be added to the package directory
    package_data={ '': ['parse_reads_cy.pyx', 'Snakefile', 'config_default.yaml'] },

    # metadata for upload to PyPI
    author = "Nils Koelling",
    author_email = "git@nk.gl",
    description = "amplicon/smMIP mapping and analysis pipeline",
    long_description = long_description,
    license = "Apache License, Version 2.0",
    keywords = "amplimap amplicon smmip mapping analysis pipeline",
    url = "https://github.com/koelling/amplimap/",
    download_url="https://github.com/koelling/amplimap/archive/v%s.tar.gz" % __version__,
    platforms=["any"],
    entry_points={
        'console_scripts': [
            'amplimap = amplimap.run:main',
            'amplimap_merge = amplimap.merge_folders:main',
            'amplimap_setup = amplimap.run_setup:main',
            #'amplimap_pileup = amplimap.pileup:main',
        ]
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
    ],

    python_requires='>=3',

    install_requires=[
        'snakemake>=3.11.2,<5.7',
        'pyyaml>=3.12,<4',
        'numpy>=1.13.1,<2',
        'biopython>=1.69,<2',
        'pandas>=0.20.3,<1',
        'interlap>=0.2.5',
        'pysam>=0.11.1,<0.14',  # pysam seem to have some issues in 0.14, so we force something in-between here
        'pyfaidx>=0.4.8.4',
        'distance>=0.1.3',
        'umi_tools>=0.5.0,<1',  # umi_tools 1 changed the API used in find_umi_groups
    ],

    setup_requires=[
        'setuptools>=18.0',
        'cython',
    ],

    ext_modules = [Extension("amplimap.parse_reads_cy", ["amplimap/parse_reads_cy.pyx"])]
)



"""
This does some basic testing of the whole pipeline. Note this needs to be run through pytest and doesn't use unittest.
"""

#make sure we can import from package directory
import sys, os, re, gzip
packagedir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, packagedir)

import shutil
import subprocess
import pathlib
import pandas as pd

import amplimap.run

#we need to build the cython file here, since snakemake will call an external python that wouldn't
#inherit pyximport
os.system("cythonize -i {}".format(os.path.join(packagedir, "amplimap", "parse_reads_cy.pyx")))

#set config
test_config_path = os.path.join(packagedir, "sample_data", "config_default.yaml")
os.environ['AMPLIMAP_CONFIG'] = test_config_path

def init_wd(path, reads_in_path, umi_one = 0, umi_two = 0, remove_analysis = True):
    assert os.path.isdir(path)

    # remove previous output
    if remove_analysis:
        shutil.rmtree(os.path.join(path, 'analysis'), ignore_errors=True)

    # remove snakemake temp
    shutil.rmtree(os.path.join(path, '.snakemake'), ignore_errors=True)

    # remove previous reads_in and then prepare a new one
    shutil.rmtree(os.path.join(path, 'reads_in'), ignore_errors=True)
    os.mkdir(os.path.join(path, 'reads_in'))

    #turn .fastq files into .fastq.gz
    for file in os.listdir(reads_in_path):
        if umi_one > 0 and '_R1' in file:
            umi_len = umi_one
        elif umi_two > 0 and '_R2' in file:
            umi_len = umi_two
        else:
            umi_len = 0

        next_umi = ''
        with open(os.path.join(reads_in_path, file), 'rt') as fin, gzip.open(os.path.join(path, 'reads_in', '{}.gz'.format(file)), 'wt') as fout:
            for ix, line in enumerate(fin):
                assert line.startswith('@') or ix % 4 != 0
                assert line.startswith('+') or ix % 4 != 2

                if umi_len > 0:
                    if ix % 4 == 0: #name
                        match = re.search(r'_UMI-([^_]+)', line)
                        assert match, 'UMI missing: %s' % line
                        next_umi = match.group(1)
                    elif ix % 4 == 1: #seq
                        line = '{}{}'.format(next_umi[0:umi_len], line)
                    elif ix % 4 == 3: #qual
                        line = '{}{}'.format('A' * umi_len, line)

                fout.write(line)

def test_version(capsys):
    assert amplimap.run.main(['--version']) == 0
    captured = capsys.readouterr()
    assert captured.out.strip() == '{} {}'.format(amplimap.run.__title__, amplimap.run.__version__)

def test_config(capsys):
    assert amplimap.run.main(['--print-config']) == 0
    captured = capsys.readouterr()
    assert 'Reading additional configuration file: {}'.format(os.path.join(packagedir, "sample_data", "config_default.yaml")) in captured.err

def test_config_env(capsys):
    extra_config_path = os.path.join(packagedir, "sample_data", "extra_config.yaml")
    os.environ['AMPLIMAP_CONFIG'] = extra_config_path
    assert amplimap.run.main(['--print-config']) == 0
    captured = capsys.readouterr()
    os.environ['AMPLIMAP_CONFIG'] = test_config_path #reset, so we don't affect later tests

    # #for debugging:
    # with capsys.disabled():
    #     sys.stdout.write(captured.err)
    #     sys.stdout.write(captured.out)

    assert 'Reading additional configuration file: {}'.format(extra_config_path) in captured.err
    assert 'aligner: star' in captured.out

def test_config_custom_caller(tmp_path, capsys):
    cfg = tmp_path / "config.yaml"
    cfg.write_text('''
variants:
  caller: custom_test
modules:
  custom_test: 'custom_test/123'
tools:
  custom_test:
    call_command: 'date'
    ''')

    extra_config_path = str(cfg)
    os.environ['AMPLIMAP_CONFIG'] = extra_config_path
    assert amplimap.run.main(['--print-config']) == 0
    captured = capsys.readouterr()
    os.environ['AMPLIMAP_CONFIG'] = test_config_path # reset, so we don't affect later tests

    assert 'Reading additional configuration file: {}'.format(extra_config_path) in captured.err
    assert 'caller: custom_test' in captured.out

def test_config_custom_module_missing_fails(tmp_path, capsys):
    cfg = tmp_path / "config.yaml"
    cfg.write_text('''
modules:
  custom_test: 'custom_test/123'
    ''')

    extra_config_path = str(cfg)
    os.environ['AMPLIMAP_CONFIG'] = extra_config_path
    assert amplimap.run.main(['--print-config']) == 1
    captured = capsys.readouterr()
    os.environ['AMPLIMAP_CONFIG'] = test_config_path # reset, so we don't affect later tests

    assert 'Reading additional configuration file: {}'.format(extra_config_path) in captured.err
    assert 'unknown or invalid settings:\n\t- modules:custom_test' in captured.err

def test_config_invalid_caller_fails(tmp_path, capsys):
    cfg = tmp_path / "config.yaml"
    cfg.write_text('''
variants:
  caller: custom_test
    ''')

    extra_config_path = str(cfg)
    os.environ['AMPLIMAP_CONFIG'] = extra_config_path
    assert amplimap.run.main(['--print-config']) == 1
    captured = capsys.readouterr()
    os.environ['AMPLIMAP_CONFIG'] = test_config_path # reset, so we don't affect later tests

    assert 'Reading additional configuration file: {}'.format(extra_config_path) in captured.err
    assert 'variants: caller must be one of' in captured.err

def test_config_env_invalid(capsys):
    extra_config_path = os.path.join(packagedir, "sample_data", "extra_config_invalid.yaml")
    os.environ['AMPLIMAP_CONFIG'] = extra_config_path
    assert amplimap.run.main(['--print-config']) == 1
    captured = capsys.readouterr()
    os.environ['AMPLIMAP_CONFIG'] = test_config_path #reset, so we don't affect later tests

    assert 'Reading additional configuration file: {}'.format(extra_config_path) in captured.err
    assert 'Your configuration file(s) contain unknown or invalid settings:' in captured.err

def check_run(capsys, wd_path, rules=['pileups'], run=True):
    #dry-run
    assert amplimap.run.main(['--working-directory={}'.format(wd_path)] + rules) == 0
    captured = capsys.readouterr()
    assert '{} {} dry run successful.'.format(amplimap.run.__title__, amplimap.run.__version__) in captured.err.strip()

    # full run
    if run:
        assert amplimap.run.main(['--working-directory={}'.format(wd_path), '--run'] + rules) == 0
        captured = capsys.readouterr()
        assert '{} {} finished!'.format(amplimap.run.__title__, amplimap.run.__version__) in captured.err.strip()

    return captured

def check_default_stats(wd_path, is_trimmed=True):
    samples = pd.read_csv(os.path.join(wd_path, 'analysis', 'reads_parsed', 'stats_samples.csv'))
    assert len(samples) == 1

    assert samples.loc[0, 'sample'] == 'S1'
    assert samples.loc[0, 'files'] == 2
    assert samples.loc[0, 'pairs_total'] == 7
    assert samples.loc[0, 'pairs_good_arms'] == 6
    assert samples.loc[0, 'pairs_r1_too_short'] == (1 if is_trimmed else 0)

def check_default_pileups(wd_path, expected_coverage = 5, include_too_short = False):
    pileups = pd.read_csv(os.path.join(wd_path, 'analysis', 'pileups', 'pileups_long.csv'))

    #we covered 11bp
    assert len(pileups) == 11

    #we should have 5 reads, except for the raw alignments where we include the pair with short r1/r2
    pileups['expected_coverage'] = expected_coverage
    if include_too_short:
        pileups.loc[(pileups.pos <= 32) | (pileups.pos >= 39), 'expected_coverage'] += 1

    #everything but 35/37 should be ref
    assert pileups.loc[~pileups.pos.isin([35,37]), 'alts'].isnull().all()
    assert (pileups.loc[~pileups.pos.isin([35,37]), 'ref_hq_count'] == pileups.loc[~pileups.pos.isin([35,37]), 'expected_coverage']).all()
    assert (pileups.loc[~pileups.pos.isin([35,37]), 'nonref_hq_count'] == 0).all()

    #only these should be nonref (36 is low-quality in one read)
    assert pileups.loc[pileups.pos.isin([35,37]), 'alts'].notnull().all()
    #one read from L001, one from L002
    assert (pileups.loc[pileups.pos == 35, 'nonref_hq_count'] == 2).all()
    assert (pileups.loc[pileups.pos == 35, 'ref_hq_count'] == pileups.loc[pileups.pos == 35, 'expected_coverage'] - 2).all()
    assert (set(pileups.loc[pileups.pos == 35, 'alts'].iloc[0].split(';')) == set(['A', 'G'])) #explicitly use iloc and no .all() here
    #just one in L001
    assert (pileups.loc[pileups.pos == 37, 'nonref_hq_count'] == 1).all()
    assert (pileups.loc[pileups.pos == 37, 'ref_hq_count'] == pileups.loc[pileups.pos == 37, 'expected_coverage'] - 1).all()
    assert (set(pileups.loc[pileups.pos == 37, 'alts'].iloc[0].split(';')) == set(['G'])) #explicitly use iloc and no .all() here


def test_multiple_input_dirs(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_multi_input")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    assert amplimap.run.main(['--working-directory={}'.format(wd_path), 'pileups']) == 1
    captured = capsys.readouterr()

    assert 'Please only provide a single input directory with all your data.' in captured.err


def test_naive_pileups(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_naive")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    check_run(capsys, wd_path)
    check_default_stats(wd_path)
    check_default_pileups(wd_path)


# NOTE: capfd instead of capsys here to also catch output from Snakemake
def test_mipgen(capfd):
    "Check that we can convert a MIPGEN table"

    wd_path = os.path.join(packagedir, "sample_data", "wd_mipgen")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    assert amplimap.run.main(['--working-directory={}'.format(wd_path), 'pileups']) == 0
    captured = capfd.readouterr()
    assert '{} {} dry run successful.'.format(amplimap.run.__title__, amplimap.run.__version__) in captured.err.strip()

    assert amplimap.run.main(['--run', '--working-directory={}'.format(wd_path), 'pileups']) == 1
    captured = capfd.readouterr()

    assert '2/3 probes left after merging probes by ID' in captured.err.strip()
    assert 'ABORTED: Did not find any read pairs with the expected primers sequences' in captured.err.strip()
    assert 'The UMI length settings are currently {} bp for read one and {} bp for read two.'.format(
        0, 0
    ) in captured.err.strip()


# NOTE: capfd instead of capsys here to also catch output from Snakemake
def test_config_umi_error(capfd):
    "Check that we get the correct error message if UMI lengths are wrong"

    os.environ['AMPLIMAP_CONFIG'] = os.path.join(packagedir, "sample_data", "config_default_wrong_umi.yaml")

    wd_path = os.path.join(packagedir, "sample_data", "wd_naive")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    assert amplimap.run.main(['--working-directory={}'.format(wd_path), 'pileups']) == 0
    captured = capfd.readouterr()
    assert '{} {} dry run successful.'.format(amplimap.run.__title__, amplimap.run.__version__) in captured.err.strip()

    assert amplimap.run.main(['--run', '--working-directory={}'.format(wd_path), 'pileups']) == 1
    captured = capfd.readouterr()

    assert 'ABORTED: Did not find any read pairs with the expected primers sequences' in captured.err.strip()
    assert 'The UMI length settings are currently {} bp for read one and {} bp for read two.'.format(
        3, 1
    ) in captured.err.strip()

    os.environ['AMPLIMAP_CONFIG'] = test_config_path  # reset, so we don't affect later tests


def test_variants(capsys):
    """
    For this test we have some intermediate results already, otherwise we would require a variant caller to be installed.
    We also skip the annotation part.
    """

    wd_path = os.path.join(packagedir, "sample_data", "special_wd_variants")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"), remove_analysis=False)

    # clean up possible old results
    for file in ['config_used.yaml', 'variants_raw/variants_merged.csv']:
        if os.path.exists(os.path.join(wd_path, 'analysis', file)):
            os.unlink(os.path.join(wd_path, 'analysis', file))

    # now update file modification time to pretend we called variants
    for file in ['targets.bed', 'targets_merged.bed', 'versions/gatk.txt', 'variants_raw/S1.vcf']:
        pathlib.Path(os.path.join(wd_path, 'analysis', file)).touch()

    rules_manual = [
        '--resume',
        os.path.join('analysis', 'variants_raw', 'variants_summary.csv'),
    ]

    # just run the variants rule, we can't run from scratch since we won't have a caller
    captured = check_run(capsys, wd_path, rules = rules_manual, run=False)
    # make sure we are not trying to rerun everything
    # NOTE: this will contain output from above, so we can't fail on align_pe
    assert not 'call_variants_raw' in captured.out.strip()
    # make sure we want to reannotate
    assert 'variants_merge_unannotated' in captured.out.strip()
    # now actually run
    captured = check_run(capsys, wd_path, rules = rules_manual)

    # check variant files
    variants_merged = pd.read_csv(os.path.join(wd_path, 'analysis', 'variants_raw', 'variants_merged.csv'), index_col=['Chr', 'Start'])
    assert len(variants_merged) == 5
    assert len(variants_merged.loc['U00096.3', 35]) == 1
    assert len(variants_merged.loc['U00096.3', 36]) == 1
    assert len(variants_merged.loc['U00096.3', 37]) == 1
    assert len(variants_merged.loc['U00096.3', 45]) == 2

    variants_summary = pd.read_csv(os.path.join(wd_path, 'analysis', 'variants_raw', 'variants_summary.csv'), index_col=['Chr', 'Start', 'Alt'])
    assert len(variants_summary) == 5

    assert variants_summary.loc['U00096.3', 35, 'C']['Ref'] == 'T'
    assert variants_summary.loc['U00096.3', 36, 'A']['Ref'] == 'C'
    assert variants_summary.loc['U00096.3', 37, 'T']['Ref'] == 'TGTG'
    assert variants_summary.loc['U00096.3', 45, 'G']['Ref'] == 'T'
    assert variants_summary.loc['U00096.3', 45, 'C']['Ref'] == 'T'

    assert variants_summary.loc['U00096.3', 35, 'C']['Var_Zygosity'] == 'Het'
    assert variants_summary.loc['U00096.3', 36, 'A']['Var_Zygosity'] == 'HOM'
    assert variants_summary.loc['U00096.3', 37, 'T']['Var_Zygosity'] == 'Het'
    assert variants_summary.loc['U00096.3', 45, 'G']['Var_Zygosity'] == 'REF'
    assert variants_summary.loc['U00096.3', 45, 'C']['Var_Zygosity'] == 'Het'

    assert variants_summary.loc['U00096.3', 35, 'C']['Target'] == 'target_0'
    assert variants_summary.loc['U00096.3', 36, 'A']['Target'] == 'target_0'
    assert variants_summary.loc['U00096.3', 37, 'T']['Target'] == 'target_0'
    assert pd.isna(variants_summary.loc['U00096.3', 45, 'G']['Target'])
    assert pd.isna(variants_summary.loc['U00096.3', 45, 'C']['Target'])

    assert variants_summary.loc['U00096.3', 35, 'C']['Var_FailedFilters'] == 'badReads'
    assert variants_summary.loc['U00096.3', 45, 'G']['Var_FailedFilters'] == 'badReads'
    assert variants_summary.loc['U00096.3', 45, 'C']['Var_FailedFilters'] == 'badReads'
    assert variants_summary['Var_FailedFilters'].isnull().sum() == 2


def test_naive_pileups_notrim(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_naive_notrim")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    check_run(capsys, wd_path)
    check_default_stats(wd_path, is_trimmed = False)
    #check_default_pileups(wd_path, expected_coverage = 6)


def test_bwa_pileups(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_bwa")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    check_run(capsys, wd_path)
    check_default_stats(wd_path)
    check_default_pileups(wd_path)


def test_bwa_pileups_notrim(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_bwa_notrim")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    check_run(capsys, wd_path)
    check_default_stats(wd_path, is_trimmed = False)
    #check_default_pileups(wd_path, expected_coverage = 6)


def test_raw_read_pileups(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_bwa_raw")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    check_run(capsys, wd_path)

    #check custom stats
    samples = pd.read_csv(os.path.join(wd_path, 'analysis', 'reads_parsed', 'stats_samples.csv'))
    assert len(samples) == 1

    assert samples.loc[0, 'sample'] == 'S1'
    assert samples.loc[0, 'files'] == 2
    assert samples.loc[0, 'pairs_total'] == 7
    assert samples.loc[0, 'pairs_good_arms'] == 7
    assert samples.loc[0, 'pairs_r1_too_short'] == 0

    check_default_pileups(wd_path, include_too_short = True)


def test_umi_pileups(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_umis")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"), umi_one = 3, umi_two = 4)

    check_run(capsys, wd_path)
    check_default_stats(wd_path)
    check_default_pileups(wd_path, expected_coverage=4) #one less, because two read pairs have same umi!

    #check umi-specific stats
    stats_reads = pd.read_csv(os.path.join(wd_path, 'analysis', 'reads_parsed', 'stats_reads.csv'))
    assert len(stats_reads) == 1
    assert stats_reads.loc[0, 'sample'] == 'S1'
    assert stats_reads.loc[0, 'probe'] == 'Probe1'
    assert stats_reads.loc[0, 'read_pairs'] == 6
    assert stats_reads.loc[0, 'umis_total'] == 5
    assert stats_reads.loc[0, 'umis_coverage_max'] == 2


def test_umi_dedup(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_umis")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"), umi_one = 3, umi_two = 4)

    check_run(capsys, wd_path, rules = ['dedup_bams'])

    import pysam
    #before dedup we had five aligned read pairs
    n_in = pysam.AlignmentFile(os.path.join(wd_path, 'analysis', 'bams', 'S1.bam')).count(until_eof=True)
    assert n_in == 5 * 2

    #after dedup we have four (two read pairs have same UMI)
    n_dedup = pysam.AlignmentFile(os.path.join(wd_path, 'analysis', 'bams_umi_dedup', 'S1.bam')).count(until_eof=True)
    assert n_dedup == 4 * 2


def test_naive_pileups_simulation(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_naive")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))
    shutil.rmtree(os.path.join(wd_path, 'test__GGCAATATGT_GGCAATCTGT_100'), ignore_errors=True) #make sure this doesn't exist

    #run simulation, replacing A@30 to C
    assert amplimap.run.main(['--working-directory={}'.format(wd_path), 'test__GGCAATATGT_GGCAATCTGT_100/test_pileups.done', '--run']) == 0
    captured = capsys.readouterr()
    assert '{} {} finished!'.format(amplimap.run.__title__, amplimap.run.__version__) in captured.err.strip()

    pileups = pd.read_csv(os.path.join(wd_path, 'test__GGCAATATGT_GGCAATCTGT_100', 'pileups', 'pileups_long.csv'))
    assert len(pileups) == 11

    #we should have an A>C SNP at pos 30, in addition to the others
    assert pileups.loc[~pileups.pos.isin([30, 35, 37]), 'alts'].isnull().all()
    assert (pileups.loc[pileups.pos == 30, 'nonref_hq_count'] == 5).all()
    assert (pileups.loc[pileups.pos == 30, 'ref_hq_count'] == 0).all()
    assert (set(pileups.loc[pileups.pos == 30, 'alts'].iloc[0].split(';')) == set(['C'])) #explicitly use iloc and no .all() here


from cpython.mem cimport PyMem_Malloc, PyMem_Free

#check in jupyter with: %%cython --annotate

cdef unsigned int mismatch_count_with_max_cy(unsigned int my_len, char* a, char* b, unsigned int max_mismatches = 1):
    """Count number of mismatches between two strings, but stop once max_mismatches has been reached and just return max_mismatches+1."""
    cdef unsigned int found_mm = 0

    cdef unsigned int i
    for i in range(my_len):
        if a[i] != b[i] and a[i] != b'.' and b[i] != b'.':
            found_mm += 1
            if found_mm > max_mismatches:
                return found_mm

    return found_mm

#a is full string, b is prefix
def mismatch_count_with_max(a, b, max_mismatches = 1):
    """Python wrapper for mismatch_count_with_max_cy."""
    comparable_len = min(len(a), len(b))
    found_mm = mismatch_count_with_max_cy(comparable_len, a[:comparable_len], b[:comparable_len], max_mismatches)

    #if len(a) > len(b) [normal case, where b is prefix of a] we just ignore the rest
    #if len(b) > len(a) we add to mismatch
    if len(b) > len(a):
        found_mm += len(b) - len(a)

    return found_mm

cdef list find_prefixes_with_mismatches_remove(char* str, tuple p, unsigned int k):
    """Search for overlap between a single string and a list (tuple) of prefix strings at the same time.

    Args:
        str: The string to search for (eg. read).
        p: Tuple of prefix strings (eg. expected primer arms).
        k: Maximum number of mismatches to allow.

    Returns:
        list of tuples (prefix index, number of mismatches)
    """
    cdef char str_char
    cdef char prefix_char
    cdef unsigned int char_ix
    cdef unsigned int prefix_ix
    cdef unsigned int n = len(p)
    cdef unsigned int max_search_len = 0
    cdef unsigned int str_len = len(str)
    cdef bytes prefix_str
        
    result = []
    
    #TODO: these should really be allocated outside the main loop already! at the moment we realloc this for every read...
    cdef char **prefixes = <char **>PyMem_Malloc(n * sizeof(char *))
    cdef unsigned int *lengths = <unsigned int *>PyMem_Malloc(n * sizeof(unsigned int))
    cdef unsigned int *mismatches = <unsigned int *>PyMem_Malloc(n * sizeof(unsigned int))
    if not prefixes or not mismatches or not lengths:
        raise MemoryError()
    try:
        for prefix_ix in range(n):
            prefix_str = p[prefix_ix]
            prefixes[prefix_ix] = prefix_str
            lengths[prefix_ix] = len(prefix_str)
            mismatches[prefix_ix] = 0
            
            if lengths[prefix_ix] > max_search_len:
                max_search_len = lengths[prefix_ix]
                
        if max_search_len > str_len:
            max_search_len = str_len

        #print('max len =', max_search_len)
        for char_ix in range(max_search_len):
            str_char = str[char_ix]
            #print('#', char_ix, ':', str_char)

            for prefix_ix in range(n):
                if mismatches[prefix_ix] <= k:
                    #have we reached the end?
                    if lengths[prefix_ix] <= char_ix:
                        #then this as a match
                        #print('found: ', prefix_ix, 'with mm =', mismatches[prefix_ix], 'after', char_ix, 'chars')

                        result.append((prefix_ix, mismatches[prefix_ix]))
                        mismatches[prefix_ix] = k+1
                    else:
                        prefix_char = prefixes[prefix_ix][char_ix]
                        #print(prefix_char)
                        #do we have a mismatch
                        if str_char != prefix_char and str_char != b'.' and prefix_char != b'.':
                            mismatches[prefix_ix] += 1

        #do one last loop through prefixes to find full-length cases (where lengths[prefix_ix] >= max_search_len)
        for prefix_ix in range(n):
            #cases where search len < prefix len are never a match, since search string was too short
            #NOTE: may want to change this by instead adding length diff to mismatch count, but that would be inconsistent with how we handle R2
            if lengths[prefix_ix] <= max_search_len:
                if mismatches[prefix_ix] <= k:            
                    result.append((prefix_ix, mismatches[prefix_ix]))
                        
    finally:
        # return the previously allocated memory to the system
        PyMem_Free(prefixes)
        PyMem_Free(lengths)
        PyMem_Free(mismatches)
                
    return result

def find_probe_from_reads(
        tuple first_arms_probes,
        tuple first_arms_seq,
        dict second_arms,
        char* first_read_str, char* second_read_str,
        unsigned int umi_one, unsigned int umi_two,
        unsigned int mismatches,
        debug = False):
    """
    Find probes for which the primer arms match the starts of the first and second read.
    """
    cdef unsigned int match_ix, probe_ix, mismatches_first
    cdef char* second_prefix
    
    matches = []
    matches_first = find_prefixes_with_mismatches_remove(first_read_str[umi_one:], first_arms_seq, mismatches)    
    if debug:
        print('R1:', first_read_str[umi_one:])
        for fas in first_arms_seq:
            print(fas)
        print('Matches for first arm:', matches_first)

    for match_ix in range(len(matches_first)):
        probe_ix = matches_first[match_ix][0]
        mismatches_first = matches_first[match_ix][1]
        probe = first_arms_probes[probe_ix]
        second_prefix = second_arms[probe]

        if debug:
            print('R2:', second_read_str[umi_two:], 'len =', len(second_read_str) - umi_two, len(second_prefix))
            print('Testing second arm:', probe, second_prefix)

        if len(second_read_str) - umi_two >= len(second_prefix):
            mismatches_second = mismatch_count_with_max_cy(len(second_prefix), second_read_str[umi_two:], second_prefix, mismatches)  

            if debug:
                print('Mismatches in second arm:', mismatches_second)        
            if mismatches_second <= mismatches:
                matches.append([probe, mismatches_first, mismatches_second])
        elif debug:
            print('Too short!')

    return matches


# -*- coding: utf-8 -*-
import re
import umi_tools.network

extended_read_name_regex = r'^(.+):::pr_([^:]+):::umi_([^:]*)$'
def make_extended_read_name(original_name: str, probe: str, umi: str) -> str:
    """Generate read name that contains probe name and UMI."""
    assert len(original_name) > 0
    assert len(probe) > 0

    return '%s:::pr_%s:::umi_%s' % (original_name, probe, umi)

def parse_extended_read_name(extended_read_name: str) -> (str, str, str):
    """Obtain original read name, probe name and UMI from extended read name."""
    match = re.search(extended_read_name_regex, extended_read_name)
    if match:
        original_name, probe, umi = match.groups()
        return original_name, probe, umi
    else:
        raise Exception('Read name "%s" in unexpected format, does not contain probe id in read name!' % extended_read_name)

def find_umi_groups(umi_counts: dict, id_offset = 0) -> dict:
    """
    Calculate dict with UMI group ID (values) for each raw UMI (keys).

    Args:
        umi_counts (dict): Counts for each raw UMI sequence, keys must be bytes (not unicode strings).
        id_offset (int): Optional offset to add to all UMI IDs.
    """
    umi_to_group = {}

    if len(umi_counts) > 0:
        #check keys
        assert isinstance(next(iter(umi_counts.keys())), (bytes, bytearray))

        processor = umi_tools.network.UMIClusterer('directional')
        umi_groups = processor(umi_counts.keys(), umi_counts, threshold=1) #TODO: this should be --edit-distance-threshold, defaults to 1 in Utilities.py
        for umi_group_id, umi_group_members in enumerate(umi_groups):
            for raw_umi in umi_group_members:
                umi_to_group[raw_umi] = id_offset + umi_group_id

    return umi_to_group